/* --------------------------------------------------------
                       blink_tim2_int.c

     blinkende LED ueber Timer2-Interrupt gesteuert

     MCU : ch32v003

     05.04.2025  R. Seelig
   -------------------------------------------------------- */

/*

                  CH32V003 A4M6                                        CH32V003F4P6

                  +-----------+                                      +--------------+
        PC1 / sda |1    C   16| PC0                         PD4 / a7 | 1         20 | PD3 / a4
        PC2 / scl |2    H   15| Vdd                   PD5 / a5 / utx | 2    C    19 | PD2 / a3
              PC3 |3    3   14| gnd                   PD6 / a6 / urx | 3    H    18 | PD1 / swio
              PC4 |4    2   13| PA2 / osc0 / a0           nrst / PD7 | 4    3    17 | PC7 / miso
       PC6 / mosi |5    V   12| PA1 / osc1 / a1      PA1 / osc0 / a0 | 5    2    16 | PC6 / mosi
       PC7 / miso |6    0   11| nrst / PD7           PA2 / osc1 / a1 | 6    V    15 | PC5 / sck / scl
       PD1 / swio |7    0   10| PD6 / a6 / urx                   gnd | 7    0    14 | PC4 / a2
         PD4 / a7 |8    3    9| PD5 / a5 / utx                   PD0 | 8    0    13 | PC3
                  +-----------+                                  Vdd | 9    3    12 | PC2 / scl
                                                                 PC0 | 10        11 | PC1 / sda
                                                                     +--------------+

*/

#include "ch32fun.h"
#include "ch32v003_gpio.h"

#define led_init()           PC3_output_init()
#define led_set()            PC3_set()
#define led_clr()            PC3_clr()

/* --------------------------------------------------------
                         timer2_init

     initialisiert Timer2 fuer einen Compare-Match-
     Betrieb im Interruptmodus.

     Uebergabe:

     prescale  : Taktteiler des Corecodes zum Zaehlers.
     cmp_match : Compare-Match (hier Autoreloadwert) bei
                 dem der Zaehler ueberlaeuft und einen
                 Interrupt ausloest

                 Bsp.: 48 MHz Coretakt, Prescale 48000
                 bedeutet 1 KHz Timer Takt = 1 ms
                 cmp_match 999 bedeutet 0..999 = 1000 Takte
                 bis zum Interruptintervall
   -------------------------------------------------------- */
void timer2_init(uint16_t prescale, uint16_t cmp_match)
{
  RCC->APB1PCENR |= (RCC_APB1Periph_TIM2);
  TIM2->CTLR1 |= TIM_CKD_DIV1;

  TIM2->PSC= prescale;
  TIM2->ATRLR= cmp_match;

  TIM2->SWEVGR |= TIM_UG;
  TIM2->DMAINTENR |=(TIM_IT_Update);
  TIM2->CTLR1 |= TIM_CEN;

  NVIC_SetPriority(TIM2_IRQn,5);
  NVIC_EnableIRQ(TIM2_IRQn);
}


/* --------------------------------------------------------
                    TIM2_IRQHandler

      Code nach TIM_IT_UPDATE wird hier mit der Anzahl
      Mikrosekunden abgearbeitet, die in cmp_match des
      timer2_init angegeben ist
   -------------------------------------------------------- */
void TIM2_IRQHandler() __attribute__( ( interrupt() ) );

void TIM2_IRQHandler()
{
  static int toggle = 0;

  toggle = toggle ^ 1;
  if (toggle) { led_set(); } else { led_clr(); }

  TIM2->INTFR &=~(uint16_t)TIM_IT_Update;
}


/* --------------------------------------------------------
                             main
   -------------------------------------------------------- */
int main(void)
{
  SystemInit();

  led_init();

  timer2_init(48000,500);

  while(1);            // Endlosschleife, Blinken geschieht im Interrupt
}
