/* -----------------------------------------------------
                        i2c_devices_soft.c

    Softwaremodul fuer rudimentaeres Ansprechen der
    I2C-Devices:

         - DS1307  (real time clock)
         - RDA5807 (UKW receiver)
         - LM75    (Temperatursensor)
         - 24LCxx  (EEProm)
         - SSD1306 (OLED-Display)

    Hardware  : CH32V003

    25.04.2025   R. Seelig
  ------------------------------------------------------ */

#include "i2c_devices_soft.h"

uint8_t eep_addr      = 0xa0;
uint8_t eep_pagesize  = 0x10;
uint8_t eep_cycletime = 7;

extern const unsigned char font8x8h[][8];


/* #################################################################
     Funktionen des DS1307 (RTC - Echtzeituhr)
   ################################################################# */

/* --------------------------------------------------
     rtc_read

     liest einen einzelnen Wert aus dem RTC-Chip

     Uebergabe:
         addr : Registeradresse des DS1307 der
                gelesen werden soll
   -------------------------------------------------- */
uint8_t rtc_read(uint8_t addr)
{
  uint8_t value;

  i2c_sendstart();
  i2c_write(rtc_addr);
  i2c_write(addr);
  i2c_stop();
  i2c_sendstart();
  i2c_write(rtc_addr | 1);
  value= i2c_read_nack();
  i2c_stop();

  return value;
}

/* --------------------------------------------------
     rtc_write

     schreibt einen einzelnen Wert aus dem RTC-Chip

     Uebergabe:
         addr : Registeradresse des DS1307 der
                geschrieben werden soll
   -------------------------------------------------- */
void rtc_write(uint8_t addr, uint8_t value)
{
  i2c_sendstart();
  i2c_write(rtc_addr);
  i2c_write(addr);
  i2c_write(value);
  i2c_stop();
}

/* --------------------------------------------------
      rtc_bcd2dez

      wandelt eine BCD Zahl (NICHT hex)  in einen
      dezimalen Wert um

      Bsp: value = 0x34
      Rueckgabe    34
   -------------------------------------------------- */
uint8_t rtc_bcd2dez(uint8_t value)
{
  uint8_t hiz,c;

  hiz= value / 16;
  c= (hiz*10)+(value & 0x0f);
  return c;
}

/* --------------------------------------------------
      rtc_getwtag

      Berechnet zu einem bestimmten Datum den
      Wochentag (nach Carl Friedrich Gauss). Der
      Function werden die Daten mittels der
      Struktur my_datum so uebergeben, wie sie
      von rtc_readdate gelesen werden. Ein Wochen-
      tag beginnt mit 0 (0 entspricht Sonntag)

      Uebergabe:
           *date  : Zeiger auf eine Struktur
                    my_datum

      Rueckgabe:
           Tag der Woche

      Bsp.:      11.04.2017   ( das ist ein Dienstag )
      Rueckgabe: 2
   -------------------------------------------------- */
uint8_t rtc_getwtag(struct my_datum *date)
{
  int tag, monat, jahr;
  int w_tag;

  tag=  rtc_bcd2dez(date->tag);
  monat= rtc_bcd2dez(date->monat);
  jahr= rtc_bcd2dez(date->jahr)+2000;

  if (monat < 3)
  {
     monat = monat + 12;
     jahr--;
  }
  w_tag = (tag+2*monat + (3*monat+3)/5 + jahr + jahr/4 - jahr/100 + jahr/400 + 1) % 7 ;
  return w_tag;
}

/* --------------------------------------------------
      rtc_readdate

      liest den DS1307 Baustein in eine Struktur
      my_datum ein.

      Rueckgabe:
          Werte der gelesenen RTC in der Struktur
          my_datum
   -------------------------------------------------- */
struct my_datum rtc_readdate(void)
{
  struct my_datum date;

  date.sek= rtc_read(0) & 0x7f;
  date.min= rtc_read(1) & 0x7f;
  date.std= rtc_read(2) & 0x3f;
  date.tag= rtc_read(4) & 0x3f;
  date.monat= rtc_read(5) & 0x1f;
  date.jahr= rtc_read(6);
  date.dow= rtc_getwtag(&date);

  return date;
}

/* --------------------------------------------------
     rtc_writedate

     schreibt die in der Struktur enthaltenen Daten
     in den RTC-Chip
   -------------------------------------------------- */
void rtc_writedate(struct my_datum *date)
{
  rtc_write(0, date->sek);
  rtc_write(1, date->min);
  rtc_write(2, date->std);
  rtc_write(4, date->tag);
  rtc_write(5, date->monat);
  rtc_write(6, date->jahr);
}

/* -----------------------------------------------------------------
     Ende DS1307
   ----------------------------------------------------------------- */


/* #################################################################
     Funktionen des LM75 Temperatursensors
   ################################################################# */

/* --------------------------------------------------
     lm75_read

     liest den Temperatursensor aus.

     Rueckgabewert:
        gelesene Temperatur x 10.

        Bsp.: Rueckgabewert von 245 entspricht 24.5 Grad Celsius
   -------------------------------------------------- */
int lm75_read(void)
{
  char       ack;
  char       t1;
  uint8_t    t2;
  int        lm75temp;

  ack= i2c_start(lm75_addr);                // LM75 Basisadresse
  if (ack)
  {

    i2c_write(0x00);                        // LM75 Registerselect: Temp. auslesen
    i2c_write(0x00);

    i2c_stop();

    ack= i2c_start(lm75_addr | 1);          // LM75 zum Lesen anwaehlen
    t1= 0;
    t1= i2c_read_ack();                     // hoeherwertigen 8 Bit
    t2= i2c_read_nack();                    // niederwertiges Bit (repraesentiert 0.5 Grad)
    i2c_stop();

  }
  else
  {
    i2c_stop();
    return -127;                            // Abbruch, Chip nicht gefunden
  }

  lm75temp= t1;
  lm75temp = lm75temp*10;
  if (t2 & 0x80) lm75temp += 5;             // wenn niederwertiges Bit gesetzt, sind das 0.5 Grad
  return lm75temp;
}

/* -----------------------------------------------------------------
     Ende LM75
   ----------------------------------------------------------------- */


/* #################################################################
     Funktionen des RDA5807 FM Receiver
   ################################################################# */

uint16_t aktfreq =   1018;      // Startfrequenz ( 101.8 MHz )
uint8_t  aktvol  =   5;         // Startlautstaerke

const uint16_t festfreq[6] = { 1018, 1048, 888, 970, 978, 999 };

uint8_t  rda5807_adrs = 0x10;   // I2C-addr. fuer sequientielllen Zugriff
uint8_t  rda5807_adrr = 0x11;   // I2C-addr. fuer wahlfreien Zugriff
uint8_t  rda5807_adrt = 0x60;   // I2C-addr. fuer TEA5767 kompatiblen Modus

uint16_t rda5807_regdef[10] ={
            0x0758,             // 00 default ID
            0x0000,             // 01 reserved
            0xF009,             // 02 DHIZ, DMUTE, BASS, POWERUPENABLE, RDS
            0x0000,             // 03 reserved
            0x1400,             // 04 SOFTMUTE
            0x84DF,             // 05 INT_MODE, SEEKTH=0110, unbekannt, Volume=15
            0x4000,             // 06 OPENMODE=01
            0x0000,             // 07 reserved
            0x0000,             // 08 reserved
            0x0000 };           // 09 reserved

uint16_t rda5807_reg[16];

/* --------------------------------------------------
     rda5807_writereg

     einzelnes Register des RDA5807 schreiben
   -------------------------------------------------- */
void rda5807_writereg(uint8_t reg)
{
  uint16_t  w;

  i2c_startaddr(rda5807_adrr,0);
  i2c_write(reg);                        // Registernummer schreiben
  w= rda5807_reg[reg];
  i2c_write16(w);         // 16 Bit Registerinhalt schreiben
  i2c_stop();
}

/* --------------------------------------------------
    rda5807_write

     alle Register es RDA5807 schreiben
   -------------------------------------------------- */
void rda5807_write(void)
{
  uint8_t   i ;

  i2c_startaddr(rda5807_adrs,0);
  for (i= 2; i< 7; i++)
  {
    i2c_write16(rda5807_reg[i]);
  }
  i2c_stop();
}

/* --------------------------------------------------
     rda5807_reset
   -------------------------------------------------- */
void rda5807_reset(void)
{
  uint8_t i;
  for (i= 0; i< 7; i++)
  {
    rda5807_reg[i]= rda5807_regdef[i];
  }
  rda5807_reg[2]= rda5807_reg[2] | 0x0002;    // Enable SoftReset
  rda5807_write();
  rda5807_reg[2]= rda5807_reg[2] & 0xFFFB;    // Disable SoftReset
}

/* --------------------------------------------------
      rda5807_poweron
   -------------------------------------------------- */
void rda5807_poweron(void)
{
  rda5807_reg[3]= rda5807_reg[3] | 0x010;   // Enable Tuning
  rda5807_reg[2]= rda5807_reg[2] | 0x001;   // Enable PowerOn

  rda5807_write();

  rda5807_reg[3]= rda5807_reg[3] & 0xFFEF;  // Disable Tuning
}

/* --------------------------------------------------
      rda5807_setfreq

      setzt angegebene Frequenz * 0.1 MHz

      Bsp.:
         rda5807_setfreq(1018);    // setzt 101.8 MHz
                                   // die neue Welle
   -------------------------------------------------- */
int rda5807_setfreq(uint16_t channel)
{

  channel -= fbandmin;
  channel&= 0x03FF;
  rda5807_reg[3]= channel * 64 + 0x10;  // Channel + TUNE-Bit + Band=00(87-108) + Space=00(100kHz)

  i2c_startaddr(rda5807_adrs,0);
  i2c_write16(0xD009);
  i2c_write16(rda5807_reg[3]);
  i2c_stop();

  delay(100);
  return 0;
}

/* --------------------------------------------------
      rda5807_setvol

      setzt Lautstaerke, zulaessige Werte fuer
      setvol 0 .. 15
   -------------------------------------------------- */
void rda5807_setvol(int setvol)
{
  rda5807_reg[5]=(rda5807_reg[5] & 0xFFF0) | setvol;
  rda5807_writereg(5);
}

/* --------------------------------------------------
      rda5807_setmono
   -------------------------------------------------- */
void rda5807_setmono(void)
{
  rda5807_reg[2]=(rda5807_reg[2] | 0x2000);
  rda5807_writereg(2);
}

/* --------------------------------------------------
      rda5807_setstero
   -------------------------------------------------- */
void rda5807_setstereo(void)
{
  rda5807_reg[2]=(rda5807_reg[2] & 0xdfff);
  rda5807_writereg(2);
}

/* -----------------------------------------------------------------
     Ende RDA5807
   ----------------------------------------------------------------- */


/* #################################################################
     Funktionen 24LCxx EEProm
   ################################################################# */

/* --------------------------------------------------
     eep_write

     schreibt einen 8-Bit Wert value an die
     Adresse adr
   -------------------------------------------------- */
void eep_write(uint16_t adr, uint8_t value)
{
  i2c_start(eep_addr);
  i2c_write16(adr);
  i2c_write(value);
  i2c_stop();
  delay(4);
}

/* --------------------------------------------------
     eep_erase

     loescht den gesamten Inhalt des EEPROMS
   -------------------------------------------------- */
void eep_erase(void)
{
  uint16_t cnt;
  uint16_t adr, len;

  adr= 0; len= 0x7fff;

  i2c_start(eep_addr);
  i2c_write16(adr);

  cnt= 0;
  do
  {
    i2c_write(0xff);
    cnt++;
    adr++;

    if ((adr % eep_pagesize == 0))                 // Pagegrenze des EEProms
    {
      i2c_stop();
      delay(4);                                    // Wartezeit bis Page geschrieben ist
      i2c_start(eep_addr);                             // neue Page oeffnen
      i2c_write16(adr);
    }
  } while (cnt< len);
  i2c_stop();
  delay(4);
}

/* --------------------------------------------------
     eep_writebuf

     schreibt mehrere Datenbytes in das EEProm

     Uebergabe:
         adr    : Adresse, ab der die Bytes im
                  EEProm gespeichert werden
         *buf   : Zeiger auf die Datenbytes, die
                  gespeichert werden sollen
         len    : Anzahl zu speichernder Bytes
   -------------------------------------------------- */
void eep_writebuf(uint16_t adr, char *buf, uint16_t len)
{
  uint16_t cnt;

  i2c_start(eep_addr);
  i2c_write16(adr);

  cnt= 0;
  do
  {
    i2c_write(*buf);
    buf++;
    cnt++;
    adr++;

    if ((adr % eep_pagesize == 0))                 // Pagegrenze des EEProms
    {
      i2c_stop();
      delay(6);                                    // Wartezeit bis Page geschrieben ist
      i2c_start(eep_addr);                         // neue Page oeffnen
      i2c_write16(adr);
    }
  } while (cnt< len);
  i2c_stop();
  i2c_stop();
  delay(6);
}

/* --------------------------------------------------
     eep_read

     liest ein einzelnes Byte aus dem EEProm an
     der Adresse adr aus
   -------------------------------------------------- */
uint8_t eep_read(uint16_t adr)
{
  uint8_t value;

  delay(1);
  i2c_start(eep_addr);                                 // sicherheitshalber
  i2c_stop();                                      // I2C Bus garantiert frei

  i2c_start(eep_addr);
  i2c_write16(adr);
  delay(5);

  i2c_start(0xa1);

  value= i2c_read_nack();
  i2c_stop();
  return value;
}

/* --------------------------------------------------
     eep_readbuf

     liest mehrere Bytes aus dem EEProm in einen
     Pufferspeicher ein.

     Uebergabe:
         adr     : Adresse, ab der im EEProm
                   gelesen wird.
         *buf    : Zeiger auf einen Speicherbereich
                   in den die Daten aus dem EEPROM
                   kopiert werden.
         len     : Anzahl der zu lesenden Bytes
   -------------------------------------------------- */
void eep_readbuf(uint16_t adr, char *buf, uint16_t len)
{
  uint16_t cnt;

  i2c_start(eep_addr);                                 // sicherheitshalber
  i2c_stop();                                      // I2C Bus garantiert frei

  i2c_start(eep_addr);
  i2c_write16(adr);
  i2c_start(0xa1);

  cnt= 0;
  do
  {
    if (len== 1)
    {
      *buf= i2c_read_nack();
      i2c_stop();
      return;
    }
    *buf= i2c_read_ack();
    buf++;
    cnt++;
    adr++;
  } while (cnt < len-1);
  i2c_read_nack();                                  // noch einmal lesen um kein ack zu schicken
  i2c_stop();                                       // und somit Adresszaehler EEPROM freigeben
  delay(1);
}

/* --------------------------------------------------
     eep_getmemsize

     Ermittelt die Speicherkapazitaet des I2C-EEProms.

     Hinweis: Es werden Speicherstellen geschrieben,
     die im Normalfall wieder restauriert werden.
     Bei einer Fehlfunktion (und einem nicht erfolgten
     Rueckschreiben) kann es zum Verlust des
     Dateninhaltes dieser Speicherstelle kommen.

     Rueckgabe:
         Speichergroesse in Bytes (max. 64 kByte)
   -------------------------------------------------- */
uint16_t eep_getmemsize(void)
{
  #define   testvalue 0xa5
  uint16_t  adr;
  uint8_t   oldbyte;

  for(adr = 0x0100; adr <= 0x8000; adr <<= 1)
  {
    oldbyte = eep_read(adr);
    eep_write(adr,testvalue);
    if( (eep_read(0x0000) == testvalue) && (eep_read(adr) == testvalue))
    {
       eep_write(adr,oldbyte);
      return adr;
    }
    else
    {
      eep_write(adr,oldbyte);
    }
  }
  return 0xffff;
}

/* -----------------------------------------------------------------
     Ende 24LCxx  EEProm
   ----------------------------------------------------------------- */


/* #################################################################
     Funktionen PCF8574 I/O-Expander
   ################################################################# */

void pcf8574_write(uint8_t value)
{
  i2c_start(pcf8574_addr);
  i2c_write(value);
  i2c_stop();
}

uint8_t pcf8574_read(void)
{
  uint8_t b;

  pcf8574_write(0xff);
  i2c_start(pcf8574_addr | 1);                  // zum lesen
  b= i2c_read_nack();
  i2c_stop();
  return b;
}

/* #################################################################
     Funktionen SSD1306 OLED-Display
   ################################################################# */

uint8_t aktxp= 0;
uint8_t aktyp= 0;
uint8_t doublechar;
uint8_t bkcolor= 0;
uint8_t textcolor= 1;


/* -------------------------------------------------------
                   ssd1306_writecmd

     sendet ein Kommandobyte an das Display
   ------------------------------------------------------- */
void ssd1306_writecmd(uint8_t cmd)
{
  i2c_start(ssd1306_addr);
  i2c_write(0x00);
  i2c_write(cmd);
  i2c_stop();
}

/* -------------------------------------------------------
                   ssd1306_writedata

    sendet ein Datenbyte an das Display
   ------------------------------------------------------- */
void ssd1306_writedata(uint8_t data)
{
  i2c_start(ssd1306_addr);
  i2c_write(0x40);
  i2c_write(data);
  i2c_stop();
}

/* -------------------------------------------------------
                     ssd1306_init

    Initialisiert das Display fuer den Gebrauch
   ------------------------------------------------------- */
void ssd1306_init(void)
{
  i2c_master_init();

  //Init LCD
  ssd1306_writecmd(0x8d);      // Ladungspumpe an
  ssd1306_writecmd(0x14);
  ssd1306_writecmd(0xaf);      // Display an
  delay(150);
  ssd1306_writecmd(0xa1);      // Segment Map
  ssd1306_writecmd(0xc0);      // Direction Map
}

/*  ---------------------------------------------------------
                              gotoxy

     legt die naechste Textausgabeposition auf dem
     Display fest. Koordinaten 0,0 bezeichnet linke obere
     Position
    --------------------------------------------------------- */
void gotoxy(uint8_t x, uint8_t y)
{
  aktxp= x;
  aktyp= y;
  x *= 8;
  y= 7-y;

  i2c_start(ssd1306_addr);
  i2c_write(0x00);

  i2c_write(0xb0 | (y & 0x0f));
  i2c_write(0x10 | (x >> 4 & 0x0f));
  i2c_write(x & 0x0f);

  i2c_stop();
}

/*  ---------------------------------------------------------
                           clrscr

      loescht den Displayinhalt mit der in bkcolor ange-
      gebenen "Farbe" (0 = schwarz, 1 = hell)
    --------------------------------------------------------- */
void clrscr(void)
{
  uint8_t x,y;

  gotoxy(0,0);
  delay(6);

  for (y= 0; y< 8; y++)                  // ein Byte in Y-Achse = 8 Pixel...
                                         // 8*8Pixel = 64 Y-Reihen
  {
    i2c_start(ssd1306_addr);
    i2c_write(0x00);

    i2c_write(0xb0 | y);            // Pageadresse schreiben
    i2c_write(0x00);                // MSB X-Adresse
    i2c_write(0x00);                // LSB X-Adresse (+Offset)

    i2c_stop();

    i2c_start(ssd1306_addr);
    i2c_write(0x40);
    for (x= 0; x< 128; x++)
    {

      if (bkcolor) i2c_write(0xff); else i2c_write(0x00);

    }
    i2c_stop();

  }
  gotoxy(0,0);
}

/*  ---------------------------------------------------------
                         oled_putchar

     gibt ein Zeichen auf dem Display aus. Steuerzeichen
     (fuer bspw. printf) sind implementiert:

             13 = carriage return
             10 = line feed
              8 = delete last char
    --------------------------------------------------------- */
void oled_putchar(uint8_t ch)
{
  uint8_t  i, b;
  uint8_t  z1;
  uint16_t z2[8];
  uint16_t z;

  if (ch== 0) return;

  if (ch== 13)                                          // Fuer <printf> "/r" Implementation
  {
    aktxp= 0;
    gotoxy(aktxp, aktyp);
    return;
  }
  if (ch== 10)                                          // fuer <printf> "/n" Implementation
  {
    aktyp++;
    if (doublechar) aktyp++;
    gotoxy(aktxp, aktyp);
    return;
  }

  if (ch== 8)
  {
    if ((aktxp> 0))
    {

      aktxp--;
      gotoxy(aktxp, aktyp);

      i2c_start(ssd1306_addr);
      i2c_write(0x40);
      for (i= 0; i< 8; i++)
      {
       if ((!textcolor)) i2c_write(0xff); else i2c_write(0x00);
      }
      i2c_stop();
      gotoxy(aktxp, aktyp);
    }
    return;

  }

  if (doublechar)
  {
    for (i= 0; i< 8; i++)
    {
      // Zeichen auf ein 16x16 Zeichen vergroessern
      z1= font8x8h[ch-' '][i];
      z2[i]= 0;
      for (b= 0; b< 8; b++)
      {
        if (z1 & (1 << b))
        {
          z2[i] |= (1 << (b*2));
          z2[i] |= (1 << ((b*2)+1));
        }
      }
    }

    i2c_start(ssd1306_addr);
    i2c_write(0x40);
    for (i= 0; i< 8; i++)
    {
      z= z2[i];
      if ((!textcolor)) z= ~z;
      z= z >> 8;
      i2c_write(z);
      i2c_write(z);
    }
    i2c_stop();
    gotoxy(aktxp, aktyp+1);

    i2c_start(ssd1306_addr);
    i2c_write(0x40);
    for (i= 0; i< 8; i++)
    {
      z= z2[i];
      if ((!textcolor)) z= ~z;
      z= z & 0xff;
      i2c_write(z);
      i2c_write(z);
    }
    i2c_stop();

    aktyp--;
    aktxp +=2;
    if (aktxp> 15)
    {
      aktxp= 0;
      aktyp +=2;
    }
    gotoxy(aktxp,aktyp);
  }
  else
  {

    i2c_start(ssd1306_addr);
    i2c_write(0x40);
    for (i= 0; i< 8; i++)
    {
      if ((!textcolor)) i2c_write(~font8x8h[ch-' '][i]);
                   else i2c_write(font8x8h[ch-' '][i]);
    }
    i2c_stop();
    aktxp++;
    if (aktxp> 15)
    {
      aktxp= 0;
      aktyp++;
    }
    gotoxy(aktxp,aktyp);
  }
}

/* -----------------------------------------------------------------
     Ende SSD1306 OLED Display
   ----------------------------------------------------------------- */
   

uint8_t ht16k_brightness      = 15;          // Helligkeit, Werte 0..15 erlaubt

uint8_t fbuffer[bufsize];

/* #################################################################
     Funktionen HT16K33
   ################################################################# */


/* --------------------------------------------------------
                       ht16k_wrcmd

     schreibt einen einzelnen Wert an die Adresse des
     Displaytreibers

     Uebergabe:
       val : zu schreibender Wert

     Rueckgabe:
       Acknowledge bei erfolgreichem Schreiben
   -------------------------------------------------------- */
uint8_t ht16k_wrcmd(uint8_t val)
{
  uint8_t ack;

  i2c_start(ht16k_addr);
  ack= i2c_write(val);
  i2c_stop();
  return ack;
}

/* --------------------------------------------------------
                       ht16k_wr2ram

     schreibt ein einzelnes Byte 'val' an die Display-
     RAM-Adresse 'segaddr'

     Uebergabe:
        val     : zu schreibender Wert
        segaddr : Adresse, an der der Wert in 'val'
                  geschickt wird.

     Rueckgabe:
        Acknowledge
   -------------------------------------------------------- */
uint8_t ht16k_wr2ram(uint8_t val, uint8_t segaddr)
{
  uint8_t ack;

  i2c_start(ht16k_addr);
  i2c_write(0x00 | segaddr);

  ack= i2c_write(val);
  i2c_stop();

  return ack;
}

/* --------------------------------------------------------
                       ht16k_wrbuffer

     schreibt den Inhalt des Framebuffers in das RAM des
     Displays.

     Da eine Commonleitung des Treibers 16 ROW-Leitungen
     betreibt, besteht ein Digit einer Commonleitung aus
     16-Bits = 2 Bytes. Aus diesem Grund belegt die
     die Adresse eines Digits im Treiber immer 2 Byte,
     so dass fuer ein 7-Segment Display immer 2
     Adressen benoetigt werden, wobei das hoeherwertige
     Byte in Verbindung mit einer max. 8 stelligen
     Anzeige somit unbenutzt ist.

     ToDo: Framebuffer auf 16-Byte erhoehen und die
     Software hier so gestalten, dass beim Schreiben
     des Buffers weitere 8 Digits an die Row-Leitungen
     8..15 angeschlossen werden koennen
   -------------------------------------------------------- */
void ht16k_wrbuffer(void)
{
  uint8_t i;

  for (i= 0; i< 8; i++)
  {
    ht16k_wr2ram(fbuffer[i], i << 1);
  }
}

/* --------------------------------------------------------
                       ht16k_init

     initialisiert den Displaytreiber
   -------------------------------------------------------- */
void ht16k_init(void)
{
  uint8_t i;

  ht16k_wrcmd(0x21);                                // Wakeup 0x20, "Normalmode" 0x01(Oszillator an)
  ht16k_wrcmd(0x81);                                // Display 0x80, "anschalten" 0x01
  ht16k_wrcmd(0xa0);                                // Row/Int-Set 0xa0, INT as row driver 0x00
  ht16k_wrcmd(0xe0 | ht16k_brightness);

  // alle LED's loeschen
  for (i= 0; i<16; i++)
  {
    ht16k_wr2ram(0, i);
  }
}

/*  ---------------------------------------------------------
                          ht16k_setdez

      zeigt einen 32-Bit Integerwert dezimal auf der
      8-stelligen Anzeige an (beschreiben des Framebuffers
      und diesen anzeigen).

      Uebergabe:
         value  : anzuzeigender Wert
         pos    : Position, ab der der Wert rechtsbuendig
                  angezeigt wird
         nozero : 1 = fuehrende Nullen werden nicht angezeigt
                  0 = werden angezeigt
    --------------------------------------------------------- */
void ht16k_setdez(int32_t value, uint8_t pos, uint8_t nozero)
{
  int32_t teiler = 10000000;
  uint8_t i, v, b;

  memset(fbuffer, 0x00, bufsize);

  for (i= 0; i< pos; i++) teiler /= 10;
  for (i= 0; i< 8-pos; i++)
  {
    v= value / teiler;
    value= value - ( v * teiler);
    teiler /= 10;
    if ((i== (8-pos-1) && nozero)) nozero= 0;
    if (v || !nozero)
    {
      b= bmps7asc[v + zifofs];            // Bitmap der Ziffer aus Asciibitmap holen
                                          // ( zifofs ist Offset zu Ascii 0, da die ersten 32
                                          // Asciizeichen nicht implementiert sind
      fbuffer[7-i]= b;

      nozero=  0;
    }
  }
  ht16k_wrbuffer();
}

/*  ---------------------------------------------------------
                        ht16k_setdp

      setzt oder loescht einen Dezimalpunkt

      Uebergabe:
         pos    : Position, an der ein Dezimalpunkt gesetzt
                  oder geloescht wird
         enable : 1 = dp wird gesetzt
                  0 = dp wird geloescht
    --------------------------------------------------------- */
void ht16k_setdp(uint8_t pos, uint8_t enable)
{

  if (enable)
  {
    fbuffer[pos] |= (0x80);
  }
  else
  {
    fbuffer[pos] &= ~(0x80);
  }
  ht16k_wrbuffer();
}

/* -----------------------------------------------------------------
     Ende HT16K33
   ----------------------------------------------------------------- */

/*
    Segmentbelegung der Anzeige:

        a
       ---
    f | g | b            Segment |  a  |  b  |  c  |  d  |  e  |  f  |  g  | Doppelpunkt (nur fuer POS1) |
       ---               ---------------------------------------------------------------------------------
    e |   | c            Bit-Nr. |  0  |  1  |  2  |  3  |  4  |  5  |  6  |              7              |
       ---
        d
*/

// Bitmapmuster ASCII - Code, beginnend mit 0x20
const uint8_t bmps7asc []  =
{
// space   !     "     #     $     %     &     '
   0x00, 0x86, 0x22, 0x7E, 0x6D, 0xD2, 0x46, 0x20,     //  0 ..  7
//   (     )     *     +     ,     -     .      /
   0x29, 0x0B, 0x21, 0x70, 0x10, 0x40, 0x80, 0x52,     //  8 .. 15
//   0     1     2     3     4     5     6     7
   0x3F, 0x06, 0x5B, 0x4F, 0x66, 0x6D, 0x7D, 0x07,     // 16 .. 23
//   8     9     :     ;     <     =     >     ?
   0x7F, 0x6F, 0x09, 0x0D, 0x61, 0x48, 0x43, 0xD3,     // 24 .. 31
 //  @     A     B     C     D     E     F     G
   0x5F, 0x77, 0x7C, 0x39, 0x5E, 0x79, 0x71, 0x3D,     // 32 .. 39
//   H     I     J     K     L     M     N     O
   0x76, 0x30, 0x1E, 0x75, 0x38, 0x15, 0x37, 0x3F,     // 40 .. 47
//   P     Q     R     S     T     U     V     W
   0x73, 0x6B, 0x33, 0x6D, 0x78, 0x3E, 0x3E, 0x2A,     // 48 .. 55
//   X     Y     Z     [     \     ]     ^     _
   0x76, 0x6E, 0x5B, 0x39, 0x64, 0x0F, 0x23, 0x08,     // 56 .. 63
//    `    a     b     c     d     e     f     g
   0x02, 0x5F, 0x7C, 0x58, 0x5E, 0x7B, 0x71, 0x6F,     // 64 .. 71
//   h     i     j     k     l     m     n     o
   0x74, 0x10, 0x0C, 0x75, 0x30, 0x14, 0x54, 0x5C,     // 72 .. 79
//  p     q     r     s     t     u     v     w
   0x73, 0x67, 0x50, 0x6D, 0x78, 0x1C, 0x1C, 0x14,     // 80 .. 87
//  x     y     z     {     |     }     ~    blk
   0x76, 0x6E, 0x5B, 0x46, 0x30, 0x70, 0x01, 0x00,     // 88 .. 95
//  o
   0x63                                                // 96 ..
};

   

const unsigned char font8x8h[][8] =
{
  {0x00,0x00,0x00,0x00,0x00,0x00,0x00,0x00},       // Ascii 32 = ' '
  {0x00,0x00,0x00,0xfa,0xfa,0x00,0x00,0x00},       // Ascii 33 = '!'
  {0x00,0xe0,0xe0,0x00,0xe0,0xe0,0x00,0x00},       // Ascii 34 = '"'
  {0x28,0xfe,0xfe,0x28,0xfe,0xfe,0x28,0x00},       // Ascii 35 = '#'
  {0x00,0x24,0x54,0xfe,0xfe,0x54,0x48,0x00},       // Ascii 36 = '$'
  {0x00,0x62,0x66,0x0c,0x18,0x30,0x66,0x46},       // Ascii 37 = '%'
  {0x00,0x0c,0x5e,0xf2,0xba,0xec,0x5e,0x12},       // Ascii 38 = '&'
  {0x00,0x00,0x20,0xe0,0xc0,0x00,0x00,0x00},       // Ascii 39 = '''
  {0x00,0x00,0x38,0x7c,0xc6,0x82,0x00,0x00},       // Ascii 40 = '('
  {0x00,0x00,0x82,0xc6,0x7c,0x38,0x00,0x00},       // Ascii 41 = ')'
  {0x10,0x54,0x7c,0x38,0x38,0x7c,0x54,0x10},       // Ascii 42 = '*'
  {0x00,0x10,0x10,0x7c,0x7c,0x10,0x10,0x00},       // Ascii 43 = '+'
  {0x00,0x00,0x00,0x06,0x06,0x00,0x00,0x00},       // Ascii 44 = ','
  {0x00,0x10,0x10,0x10,0x10,0x10,0x10,0x00},       // Ascii 45 = '-'
  {0x00,0x00,0x00,0x06,0x06,0x00,0x00,0x00},       // Ascii 46 = '.'
  {0x06,0x0c,0x18,0x30,0x60,0xc0,0x80,0x00},       // Ascii 47 = '/'
  {0x7c,0xfe,0x8a,0x92,0xa2,0xfe,0x7c,0x00},       // Ascii 48 = '0'
  {0x00,0x00,0x40,0xfe,0xfe,0x00,0x00,0x00},       // Ascii 49 = '1'
  {0x00,0x42,0xc6,0x8e,0x9a,0xf2,0x62,0x00},       // Ascii 50 = '2'
  {0x00,0x44,0xc6,0x92,0x92,0xfe,0x6c,0x00},       // Ascii 51 = '3'
  {0x18,0x38,0x68,0xc8,0xfe,0xfe,0x08,0x00},       // Ascii 52 = '4'
  {0x00,0xe4,0xe6,0xa2,0xa2,0xbe,0x9c,0x00},       // Ascii 53 = '5'
  {0x00,0x7c,0xfe,0x92,0x92,0xde,0x4c,0x00},       // Ascii 54 = '6'
  {0x00,0x80,0x80,0x8e,0x9e,0xf0,0xe0,0x00},       // Ascii 55 = '7'
  {0x00,0x6c,0xfe,0x92,0x92,0xfe,0x6c,0x00},       // Ascii 56 = '8'
  {0x00,0x60,0xf2,0x96,0x9c,0xf8,0x70,0x00},       // Ascii 57 = '9'
  {0x00,0x00,0x00,0x36,0x36,0x00,0x00,0x00},       // Ascii 58 = ':'
  {0x00,0x00,0x80,0xb6,0x36,0x00,0x00,0x00},       // Ascii 59 = ';'
  {0x00,0x10,0x38,0x6c,0xc6,0x82,0x00,0x00},       // Ascii 60 = '<'
  {0x00,0x24,0x24,0x24,0x24,0x24,0x24,0x00},       // Ascii 61 = '='
  {0x00,0x82,0xc6,0x6c,0x38,0x10,0x00,0x00},       // Ascii 62 = '>'
  {0x00,0x40,0xc0,0x8a,0x9a,0xf0,0x60,0x00},       // Ascii 63 = '?'
  {0x7c,0xfe,0x82,0xba,0xba,0xfa,0x78,0x00},       // Ascii 64 = '@'
  {0x3e,0x7e,0xc8,0x88,0xc8,0x7e,0x3e,0x00},       // Ascii 65 = 'A'
  {0xfe,0xfe,0x92,0x92,0x92,0xfe,0x6c,0x00},       // Ascii 66 = 'B'
  {0x38,0x7c,0xc6,0x82,0x82,0xc6,0x44,0x00},       // Ascii 67 = 'C'
  {0xfe,0xfe,0x82,0x82,0xc6,0x7c,0x38,0x00},       // Ascii 68 = 'D'
  {0xfe,0xfe,0x92,0x92,0x92,0x82,0x82,0x00},       // Ascii 69 = 'E'
  {0xfe,0xfe,0x90,0x90,0x90,0x80,0x80,0x00},       // Ascii 70 = 'F'
  {0x38,0x7c,0xc6,0x82,0x8a,0xce,0x4e,0x00},       // Ascii 71 = 'G'
  {0xfe,0xfe,0x10,0x10,0x10,0xfe,0xfe,0x00},       // Ascii 72 = 'H'
  {0x00,0x82,0xfe,0xfe,0x82,0x00,0x00,0x00},       // Ascii 73 = 'I'
  {0x0c,0x0e,0x02,0x02,0x02,0xfe,0xfc,0x00},       // Ascii 74 = 'J'
  {0xfe,0xfe,0x90,0xb8,0x6c,0xc6,0x82,0x00},       // Ascii 75 = 'K'
  {0xfe,0xfe,0x02,0x02,0x02,0x02,0x02,0x00},       // Ascii 76 = 'L'
  {0xfe,0xfe,0x60,0x30,0x60,0xfe,0xfe,0x00},       // Ascii 77 = 'M'
  {0xfe,0xfe,0x60,0x30,0x18,0xfe,0xfe,0x00},       // Ascii 78 = 'N'
  {0x38,0x7c,0xc6,0x82,0xc6,0x7c,0x38,0x00},       // Ascii 79 = 'O'
  {0xfe,0xfe,0x90,0x90,0x90,0xf0,0x60,0x00},       // Ascii 80 = 'P'
  {0x38,0x7c,0xc6,0x8a,0xcc,0x76,0x3a,0x00},       // Ascii 81 = 'Q'
  {0xfe,0xfe,0x90,0x90,0x98,0xfe,0x66,0x00},       // Ascii 82 = 'R'
  {0x64,0xf6,0x92,0x92,0x92,0xde,0x4c,0x00},       // Ascii 83 = 'S'
  {0x80,0x80,0xfe,0xfe,0x80,0x80,0x00,0x00},       // Ascii 84 = 'T'
  {0xfc,0xfe,0x02,0x02,0x02,0xfe,0xfc,0x00},       // Ascii 85 = 'U'
  {0xf8,0xfc,0x06,0x02,0x06,0xfc,0xf8,0x00},       // Ascii 86 = 'V'
  {0xfe,0xfe,0x0c,0x18,0x0c,0xfe,0xfe,0x00},       // Ascii 87 = 'W'
  {0xc6,0xee,0x38,0x10,0x38,0xee,0xc6,0x00},       // Ascii 88 = 'X'
  {0x00,0xe0,0xf0,0x1e,0x1e,0xf0,0xe0,0x00},       // Ascii 89 = 'Y'
  {0x82,0x86,0x8e,0x9a,0xb2,0xe2,0xc2,0x00},       // Ascii 90 = 'Z'
  {0x00,0x00,0xfe,0xfe,0x82,0x82,0x00,0x00},       // Ascii 91 = '['
  {0x80,0xc0,0x60,0x30,0x18,0x0c,0x06,0x00},       // Ascii 92 = '\'
  {0x00,0x00,0x82,0x82,0xfe,0xfe,0x00,0x00},       // Ascii 93 = ']'
  {0x00,0x10,0x30,0x60,0xc0,0x60,0x30,0x10},       // Ascii 94 = '^'
  {0x01,0x01,0x01,0x01,0x01,0x01,0x01,0x01},       // Ascii 95 = '_'
  {0x00,0x00,0x80,0xc0,0x60,0x20,0x00,0x00},       // Ascii 96 = '`'
  {0x04,0x2e,0x2a,0x2a,0x2a,0x3e,0x1e,0x00},       // Ascii 97 = 'a'
  {0xfe,0xfe,0x22,0x22,0x22,0x3e,0x1c,0x00},       // Ascii 98 = 'b'
  {0x1c,0x3e,0x22,0x22,0x22,0x36,0x14,0x00},       // Ascii 99 = 'c'
  {0x1c,0x3e,0x22,0x22,0x22,0xfe,0xfe,0x00},       // Ascii 100 = 'd'
  {0x1c,0x3e,0x2a,0x2a,0x2a,0x3a,0x18,0x00},       // Ascii 101 = 'e'
  {0x10,0x7e,0xfe,0x90,0x90,0xc0,0x40,0x00},       // Ascii 102 = 'f'
  {0x19,0x3d,0x25,0x25,0x25,0x3f,0x3e,0x00},       // Ascii 103 = 'g'
  {0xfe,0xfe,0x20,0x20,0x20,0x3e,0x1e,0x00},       // Ascii 104 = 'h'
  {0x00,0x00,0x00,0xbe,0xbe,0x00,0x00,0x00},       // Ascii 105 = 'i'
  {0x02,0x03,0x01,0x01,0xbf,0xbe,0x00,0x00},       // Ascii 106 = 'j'
  {0xfe,0xfe,0x08,0x08,0x1c,0x36,0x22,0x00},       // Ascii 107 = 'k'
  {0x00,0x00,0x00,0xfe,0xfe,0x00,0x00,0x00},       // Ascii 108 = 'l'
  {0x1e,0x3e,0x30,0x18,0x30,0x3e,0x1e,0x00},       // Ascii 109 = 'm'
  {0x3e,0x3e,0x20,0x20,0x20,0x3e,0x1e,0x00},       // Ascii 110 = 'n'
  {0x1c,0x3e,0x22,0x22,0x22,0x3e,0x1c,0x00},       // Ascii 111 = 'o'
  {0x3f,0x3f,0x24,0x24,0x24,0x3c,0x18,0x00},       // Ascii 112 = 'p'
  {0x18,0x3c,0x24,0x24,0x24,0x3f,0x3f,0x00},       // Ascii 113 = 'q'
  {0x3e,0x3e,0x20,0x20,0x20,0x30,0x10,0x00},       // Ascii 114 = 'r'
  {0x12,0x3a,0x2a,0x2a,0x2a,0x2e,0x04,0x00},       // Ascii 115 = 's'
  {0x20,0xfc,0xfe,0x22,0x22,0x26,0x04,0x00},       // Ascii 116 = 't'
  {0x3c,0x3e,0x02,0x02,0x02,0x3e,0x3e,0x00},       // Ascii 117 = 'u'
  {0x38,0x3c,0x06,0x02,0x06,0x3c,0x38,0x00},       // Ascii 118 = 'v'
  {0x3c,0x3e,0x06,0x0c,0x06,0x3e,0x3c,0x00},       // Ascii 119 = 'w'
  {0x22,0x36,0x1c,0x08,0x1c,0x36,0x22,0x00},       // Ascii 120 = 'x'
  {0x39,0x3d,0x05,0x05,0x05,0x3f,0x3e,0x00},       // Ascii 121 = 'y'
  {0x00,0x22,0x26,0x2e,0x3a,0x32,0x22,0x00},       // Ascii 122 = 'z'
  {0x00,0x10,0x10,0x7c,0xee,0x82,0x82,0x00},       // Ascii 123 = '{'
  {0x00,0x00,0x00,0xee,0xee,0x00,0x00,0x00},       // Ascii 124 = '|'
  {0x00,0x82,0x82,0xee,0x7c,0x10,0x10,0x00},       // Ascii 125 = '}'
  {0x00,0x40,0x80,0x80,0x40,0x40,0x80,0x00},       // Ascii 126 = '~'
  {0x0e,0x1e,0x32,0x62,0x62,0x32,0x1e,0x0e},       // Ascii 127 = ' '

  {0x3c,0x42,0x81,0x81,0x81,0x42,0x3c,0x00},       // Ascii 128 = rundes O
  {0x3c,0x42,0xbd,0xbd,0xbd,0x42,0x3c,0x00},       // Ascii 129 = ausgefuelltes =
  {0x00,0x00,0x70,0x88,0x88,0x88,0x70,0x00}        // Ascii 130 = hochgestelltes o
};
