/* -----------------------------------------------------
                        i2c_devices_soft.h

    Softwaremodul fuer rudimentaeres Ansprechen der
    I2C-Devices:

         - DS1307  (real time clock)
         - RDA5807 (UKW receiver)
         - LM75    (Temperatursensor)
         - 24LCxx  (EEProm)
         - SSD1306 (OLED-Display)

    Hardware  : CH32V003

    25.04.2025   R. Seelig
  ------------------------------------------------------ */

#ifndef in_i2c_devices
  #define in_i2c_devices

  #include <stdint.h>
  #include <string.h>
  #include "i2c_sw.h"


  // EEProm
  // --------------------------------------------------------------------------

  extern uint8_t eep_addr;
  extern uint8_t eep_pagesize;
  extern uint8_t eep_cycletime;

  void eep_write(uint16_t adr, uint8_t value);
  void eep_erase(void);
  void eep_writebuf(uint16_t adr, char *buf, uint16_t len);
  uint8_t eep_read(uint16_t adr);
  void eep_readbuf(uint16_t adr, char *buf, uint16_t len);
  uint16_t eep_getmemsize(void);

  // RTC (real time clock) DS1307
  // --------------------------------------------------------------------------

  #define rtc_addr            0xd0

  struct my_datum
  {
    uint8_t jahr;
    uint8_t monat;
    uint8_t tag;
    uint8_t dow;
    uint8_t std;
    uint8_t min;
    uint8_t sek;
  };

  uint8_t rtc_read(uint8_t addr);
  void rtc_write(uint8_t addr, uint8_t value);
  uint8_t rtc_getwtag(struct my_datum *date);
  uint8_t rtc_bcd2dez(uint8_t value);
  struct my_datum rtc_readdate(void);
  void rtc_writedate(struct my_datum *date);



  // Temperatursensor LM75
  // --------------------------------------------------------------------------

  #define lm75_addr    0x90

  int lm75_read(void);

  // UKW-Radio RDA5807
  // --------------------------------------------------------------------------

  #define fbandmin     870        // 87.0  MHz unteres Frequenzende
  #define fbandmax     1080       // 108.0 MHz oberes Frequenzende
  #define sigschwelle  87         // Schwelle ab der ein Sender als "gut empfangen" gilt


  extern uint16_t aktfreq;
  extern uint8_t  aktvol;

  extern const uint16_t festfreq[6];

  extern uint8_t  rda5807_adrs;
  extern uint8_t  rda5807_adrr;
  extern uint8_t  rda5807_adrt;

  extern uint16_t rda5807_regdef[10];
  extern uint16_t rda5807_reg[16];

  void rda5807_writereg(uint8_t reg);
  void rda5807_write(void);
  void rda5807_reset(void);
  void rda5807_poweron(void);
  int rda5807_setfreq(uint16_t channel);
  void rda5807_setvol(int setvol);
  void rda5807_setmono(void);
  void rda5807_setstereo(void);

  // SSD1306 OLED - Display
  // --------------------------------------------------------------------------

  #define ssd1306_addr          0x78

  extern uint8_t aktxp;
  extern uint8_t aktyp;
  extern uint8_t doublechar;
  extern uint8_t bkcolor;
  extern uint8_t textcolor;

  void ssd1306_writecmd(uint8_t cmd);
  void ssd1306_writedata(uint8_t data);
  void ssd1306_init(void);
  void gotoxy(uint8_t x, uint8_t y);
  void clrscr(void);
  void oled_putchar(uint8_t ch);


/* -----------------------------------------------------------------
     Ende RDA5807
   ----------------------------------------------------------------- */

  // PCF8574 I/O Expander
  // --------------------------------------------------------------------------

  #define pcf8574_addr          0x40

  void pcf8574_write(uint8_t value);
  uint8_t pcf8574_read(void);
  
/* -----------------------------------------------------------------
     Ende PCF8574
   ----------------------------------------------------------------- */
  
  // HT16K33 LED-Anzeigetreiber
  // --------------------------------------------------------------------------
  
  #define ascofs                0x20           // Asciitabelle der Bitmaps faengt mit Leerzeichen
                                               // (Ascii-Code 32d = 0x20) an
  #define zifofs                0x10           // Ziffernoffset des Codetabelle fuer Ziffern

  #define ht16k_addr            0xe0           // I2C-Adresse des 7-Segmenttreibers

  #define bufsize               8              // Bytegroesse des Displayframebuffers


  extern uint8_t fbuffer[bufsize];
  extern const uint8_t bmps7asc[];


  /* --------------------------------------------------------
                           Prototypen
     -------------------------------------------------------- */
  uint8_t ht16k_wrcmd(uint8_t val);
  uint8_t ht16k_wr2ram(uint8_t val, uint8_t segaddr);
  void ht16k_wrbuffer(void);
  uint8_t ht16k_setbrightness(uint8_t val);
  void ht16k_init(void);
  void ht16k_setdez(int32_t value, uint8_t pos, uint8_t nozero);
  void ht16k_setdp(uint8_t pos, uint8_t enable);
  
/* -----------------------------------------------------------------
     Ende HT16K33
   ----------------------------------------------------------------- */


#endif
