/* ------------------------------------------------------------------
                              i2c_explore.c

     Experimente mit I2C, der Einfachheit halber hier mittels
     Bitbanging Realisierung.

     Hinweis: das R-/W -Bit wird als Bestandteil der Adresse ange-
              sehen. Adressen werden also als 8-Bit Adresse
              verstanden). Somit ist dann ein Device 2 mal im Adressen
              raum vorhanden: Geradezahlige Adressen sind somit die
              "Leseadressen", ungeradzahlige die "Schreibadressen"


     Pinbelegung I2C
     ---------------
     siehe i2c_sw.h


     02.05.2025 R. Seelig
   ------------------------------------------------------------------ */


/*
                  CH32V003 A4M6                                        CH32V003F4P6

                  +-----------+                                      +--------------+
        PC1 / sda |1    C   16| PC0                         PD4 / a7 | 1         20 | PD3 / a4
        PC2 / scl |2    H   15| Vdd                   PD5 / a5 / utx | 2    C    19 | PD2 / a3
              PC3 |3    3   14| gnd                   PD6 / a6 / urx | 3    H    18 | PD1 / swio
              PC4 |4    2   13| PA2 / osc0 / a0           nrst / PD7 | 4    3    17 | PC7 / miso
       PC6 / mosi |5    V   12| PA1 / osc1 / a1      PA1 / osc0 / a0 | 5    2    16 | PC6 / mosi
       PC7 / miso |6    0   11| nrst / PD7           PA2 / osc1 / a1 | 6    V    15 | PC5 / sck / scl
       PD1 / swio |7    0   10| PD6 / a6 / urx                   gnd | 7    0    14 | PC4 / a2
         PD4 / a7 |8    3    9| PD5 / a5 / utx                   PD0 | 8    0    13 | PC3
                  +-----------+                                  Vdd | 9    3    12 | PC2 / scl
                                                                 PC0 | 10        11 | PC1 / sda
                                                                     +--------------+
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#include "ch32fun.h"
#include "ch32v003_gpio.h"
#include "systick.h"
#include "i2c_sw.h"
#include "i2c_devices_soft.h"
#include "uart.h"

#include "my_printf.h"

#define  printf               my_printf


enum { uart_stream, oled_stream };
uint8_t outstream = uart_stream;

/* --------------------------------------------------------
   my_putchar

   wird von my-printf / printf aufgerufen und hier muss
   eine Zeichenausgabefunktion angegeben sein, auf das
   printf einen Bytestream sendet !
   -------------------------------------------------------- */
void my_putchar(char ch)
{
  switch (outstream)
  {
    case uart_stream :  uart_putchar(ch); break;
    case oled_stream :  oled_putchar(ch); break;
    default          :  break;
  }
}



/* ########################################################
                   "Helferfunktionen"
   ######################################################## */

/* --------------------------------------------------
                      str_locase

      konvertier alle Grossbuchstaben eines  Strings
      in Kleinbuchstaben

       Uebergabe:
         str  => Ursprungstring
   -------------------------------------------------- */
void str_locase(char *str)
{
  while(*str)
  {
    if ((*str >= 'A') && (*str <= 'Z')) *str += 'a' - 'A';
    str++;
  }
}

/* --------------------------------------------------
                      str_hicase

      konvertier alle Kleinbuchstaben eines Strings
      zu Grossbuchstaben

       Uebergabe:
         str  => Ursprungstring
   -------------------------------------------------- */
void str_hicase(char *str)
{
  while(*str)
  {
    if ((*str >= 'a') && (*str <= 'z')) *str -= ('a' - 'A');
    str++;
  }
}

/* ----------------------------------------------------------
                          my_pow10

     einfache Funktion 10 hoch x
   ---------------------------------------------------------- */
int my_pow10(unsigned int x)
{
  #define my_powbase    10

  int i;
  int result = my_powbase;

  if (x== 0)
  {
    return 1;
  }
  else
  {
    for (i = 1; i < x; i++)
    {
      result *= my_powbase;
    }
    return result;
  }
}


/* ----------------------------------------------------------
                          my_atoi

                     my ascii to integer

     einfache Konvertierung eines Strings zu einem Integer-
     wert in Ermangelung eines atoi

     Uebergabe:
        *str: Zeiger, der einen im String enthaltenen
              Integerzahlenwert repraesentiert

     Rueckgabe:
        konvertierter Integerwert
   ---------------------------------------------------------- */
int my_atoi(char *str)
{
  int i, j;
  int res;
  int sign;
  int len;

  i = 0;
  res = 0;
  sign = 1;                        // Vorzeichenflag, 1 = positiv, -1 = negativ

  len= strlen(str);

  while(str[i] == ' ' || (str[i] >= 9 && str[i] <= 13))
  {
    i++;
  }

  if(str[i] == '-')
  {
    sign = -1;
    i++;
  }
  else if(str[i] == '+')
  {
    sign = 1;
    i++;
  }

  j = i-1;                                // Index auf Stringanfang

  i = len;
  while (str[i] < '0' || str[i] > '9')    // Suche nach der letzten gueltigen Ziffer
  {
    i--;
  }
  int num_end = i;

  // Konvertierung mit der niederwertigsten Stelle beginnen
  while(i > j)
  {
    if (str[i] >= '0' && str[i] <= '9')
    {
      res += my_pow10(num_end-i) * (str[i] - '0');
    } else
    {
      num_end--;                          // ueberspringen einer ungueltigen Ziffer
    }
    i--;
  }
  return(res * sign);
}

/* ----------------------------------------------------------
                          my_ahtoi

                    my ascii hex to integer

     einfache Konvertierung eines Hex-Strings zu einem
     Integerwert.

     Uebergabe:
        *str: Zeiger, der einen im String enthaltenen
              Hexadezimalwert repraesentiert
        *error: Zeiger der den Fehlercode aufnimmt

     Rueckgabe:
        error: 0 => kein Fehler
               1 => String war kein Hexadezimalwert
        konvertierter Integerwert
   ---------------------------------------------------------- */
uint32_t my_ahtoi(char *str, char *error)
{
  char     ch;
  uint32_t result;

  char     puffer[10];
  char     *ptr;

  strcpy(puffer,str);
  str_locase(puffer);
  ptr= &puffer[0];

  *error= 0;

  result= 0;
  do
  {
    ch= *ptr;
    if ((ch < '0') || ((ch > '9') && (ch < 'a')) || (ch > 'f') )
    {
      *error= 1;
      return 0xffffffff;
    }
    if (ch< 'a') { ch -= '0'; } else { ch -= 87; }
    result = (result << 4) + ch;

    ptr++;
  }  while (*ptr);

  return result;
}


/* ----------------------------------------------------------
                             strtok

     Sucht das erste (oder das naechste) 'token' aus s1
     heraus, d.h. eine Zeichenfolge, die kein Zeichen aus
     s2 enthaelt (sucht somit einen String bis zu einem
     Trennzeichen ab).

       char *strtok(char *s1, const char *s2);

     s2 definiert das Trennzeichen. strtok interpretiert
     s1 als eine Reihe von 'tokens', die durch ein oder
     mehrere Zeichen aus s2 voneinander getrennt sind.

     Falls kein 'token' gefunden wurde, ist das Ergebnis
     NULL, ansonsten schreibt strtok an die Stelle nach
     dem 'token' in s1 ein NULL-Zeichen und liefert einen
     Zeiger auf den Anfang des 'tokens.

     Nachfolgende Aufrufe von strtok mit NULL als erstem
     Argument bearbeiten den String s1 weiter. Es koennen
     dabei auch andere Trennzeichen (in s2) verwendet
     werden.
   ---------------------------------------------------------- */
char *strtok(char *s, const char *delim)
{
  static char *last;
  return strtok_r(s, delim, &last);
}

char *strtok_r(char *s, const char *delim, char **last)
{
  char *spanp;
  int c, sc;
  char *tok;

  if (s == NULL && (s = *last) == NULL) return (NULL);

  // fuehrende Trennzeichen ueberspringen
  c = *s++;
  for (spanp = (char *)delim; (sc = *spanp++) != 0;)
  {
    if (c == sc) c= *s++;
  }

  if (c == 0)       // keine Nicht-Trennzeichen
  {
    *last = NULL;
    return (NULL);
  }
  tok = s - 1;

  // nach dem Token scannen
  // String muss zwingend ein NULL-Zeichen haben. Wird dieses gelesen, wird der
  // die Funktion ebenfalls verlassen
  while(1)
  {
    c = *s++;
    spanp = (char *)delim;
    do
    {
      if ((sc = *spanp++) == c)
      {
        if (c == 0) s = NULL; else s[-1] = 0;
        *last = s;
        return (tok);
      }
    } while (sc != 0);
  }
}


/* --------------------------------------------------
                     str_delch

     loescht ein Zeichen aus einer Zeichenkette

     Uebergabe:   str     => Ursprungstring
                  pos     => Loeschposition
   -------------------------------------------------- */
void str_delch(char *str, int pos)
{
  int  l,i;
  char *hptr;

  l= strlen(str);
  hptr= str+pos;
  for (i= pos+1; i!= l+1; i++)
  {
    *hptr= *(hptr+1);
    hptr++;
  }
}

/* --------------------------------------------------
                     str_findch

      findet das erste Vorkommen eines Zeichens in
      einem String. Stringlaenge darf maximal 254
      Zeichen lang sein (da der Rueckgabewert ein
      uint8_t ist)

      Uebergabe:
        str      : Zeiger auf Quellstring
        ch       : zu findendes Zeichen
      Rueckgabe:
        Position des Zeichens, 0 wenn kein Zeichen
        gefunden wurde
   -------------------------------------------------- */
char str_findch(char *str, char ch)
{
  int pos;

  pos= 1;
  while (*str)
  {
    if (*str == ch) return pos;
    pos++;
    str++;
  }
  return 0;
}

/* --------------------------------------------------
                   str_delallch

     sucht nach dem Zeichen 'ch' und loescht dieses
     aus dem String

      Uebergabe:
        str     : Zeiger auf Quellstring
        ch      : zu loeschende Zeichen
   -------------------------------------------------- */
void str_delallch(char *str, char ch)
{
  int spos;
  int cnt= 0;

  spos= str_findch(str,ch);
  if (!spos) return;
  do
  {
    spos= str_findch(str,ch);
    if (spos)
    {
      str_delch(str,spos-1);
    }
    cnt++;
  } while((spos) && (cnt< 255));
}

/* --------------------------------------------------
                   str_replacech

     ersetzt das erste Auftreten des Zeichens src_ch
     im String str mit dem Zeichen dest_ch im String

      Uebergabe:
        str     : Zeiger auf Quellstring
        ch      : zu loeschende Zeichen
   -------------------------------------------------- */
void str_replacech(char *str, char dest_ch, char src_ch)
{
  int spos;

  spos= str_findch(str,src_ch);
  if (spos)
  {
    str += spos-1;
    *str= dest_ch;
  }
}

/* --------------------------------------------------
                      str_insert

       fuegt einen String in einen anderen ein.

       VORSICHT: Es wird keine Pruefung vorgenommen,
       ob der Bufferspeicher ausreichend ist.

       In diesem Falle hier werden lokal 100 Byte
       Pufferspeicher angelegt. Hier ist von Fall
       zu Fall zu pruefen, ob dieser ausreichend
       gross ist!

       Uebergabe:
         *dest : Zielstring, in den eingefuegt wird
         *src  : der einzufuegende String
         pos   : Position, an der eingefuegt wird
   -------------------------------------------------- */
void str_insert(char *dest, char *src, int pos)
{
  char buffer[100] = "";

  strcpy(buffer, dest+pos-1);
  strcpy(dest + pos-1, src);
  strcpy(dest + strlen(dest), buffer);
  return;
}

/* --------------------------------------------------
                      uart_lineinput

     Eingabe eines Strings in den Buffer *string

     Beim Start der Funktion wird der in *string ab-
     gelegte Text angezeigt und der Cursor an das
     Ende dieses Textes gesetzt.

     Einschraenkungen:

     Im Integer- und Kommamodus ist die Eingabe von
     Ziffern auf max. 8 begrenzt (da es sonst zu
     einem Werteueberlauf kommen kann).


     Uebergabe:

            *string   : Zeiger auf String, der zum
                        einen angezeigt wird, zum
                        anderen wird in diesen Speicher-
                        bereich auf den der Zeiger
                        zeigt der String abgelegt

            mode      : Eingabemodus
                           0 => Stringeingabe
                           1 => Integereingabe, nur
                                Ziffern von 0..9,'-'
                           2 => Kommazahleingabe
                                 wie Mode 1 mit
                                zusaetzlich erlaubtem
                                '.' Zeichen
                           3 => Hexadezimaleingabe,
                                Eingabe von 0..9, A..F

            anz       : max. Anzahl Zeichen, die
                        eingegeben werden koennen

     Rueckgabe:
       Zeiger auf den verarbeiteten String
   -------------------------------------------------- */
char *uart_lineinput(char *string, uint8_t mode, char anz)
{

  uint8_t      cnt;
  char         ch;
  char         *ptr;
  char         kptr;
  char         dp, tc;
  char         zifanz;

  dp= 0;
  if (strlen(string)> 0)
  {
    kptr = str_findch(string, '.');
    if (kptr) dp= 1;
  }

  if (mode== 3) { str_hicase(string); }

  // Ziffernanzahl fuer Integer und Komma auf 8 begrenzen, daher feststellen
  // wieviele Ziffern im Vorgabestring enthalten sind
  if ((mode== 1) || (mode== 2))
  {
    zifanz= strlen(string);
    if (zifanz > 0)
    {
      if (dp) { zifanz--; }
      if (*string == '-') { zifanz--; }
    }
//    printf("\n\rZifanz: %d\n\r",zifanz);
  }

  cnt= strlen(string);
  printf("%s", string);
  ptr= string+cnt;

  do
  {
    ch= uart_getchar();

    if ((ch!= 0x0a) && (ch != 0x0d))                // Enter Taste ?
    {

      if ((ch== 8) || (ch== 127))                   // Backspace
      {
        if (cnt> 0)                                 // ist ueberhaupt ein Zeichen vorhanden
        {
          tc= *(ptr-1);
          if (tc == '.') { dp= 0; }
          if ( !((tc == '.') || (tc == '-'))) { zifanz--; }
          ptr--;
          cnt--;
          *ptr= 0;                                  // neues Stringende

          // letzes Zeichen im Terminal loeschen
          uart_putchar(0x08);
          uart_putchar(' ');
          uart_putchar(0x08);
        }
      }
      else
      {                                                                         // Zeicheneingabe, erlaubte Zeichen in den entsprechenden
                                                                                // Modi pruefen
        if (mode==3)
        {
          if ((ch>= 'a') && (ch<= 'f')) { ch -= 32; }                           // bei hexadezimaler Eingabe Grossbuchstaben verwenden
        }


        if (  ((mode== 0) && (ch >= 32) && (ch <= 126))                         // anhaengen Ascii-Zeichen (Stringmode)

           || (( (mode== 1) && (ch >= '0') && (ch <= '9') )  || (ch== '-'))     // anhaengen Dezimalziffer und '-' Zeichen (Integermode)

           || ((mode== 2) && (((ch >= '0') && (ch <= '9')) ||                   // anhaengen Kommazahl (Floatmode)
                               (ch=='.') || (ch=='-')))

           || ((mode== 3) && ( ((ch >= '0') && (ch <= '9')) ||                  // anhaengen Hexadezimalzifer (Hexmode)
                               ((ch >= 'A') && (ch <= 'F'))) ) )

        {
          if ((cnt< anz-1) && (mode== 0) && (ch== '#'))                          // Pseudo-Enter fuer Texteingabe
          {
            *ptr= '\n';
            ptr++;
            *ptr= '\r';
            ptr++;
            cnt += 2;
            printf("\n\r");
          }
          else
          {
            if ((cnt< anz) && (!((mode==3) && (ch== '-'))) )
            {
              // '-' Zeichen nur erlauben im Integer- und Floatmode und auch nur, wenn noch kein
              // Minuszeichen angegeben wurde und nur, wenn der Cursor ganz am Anfang steht
              if (!( ((mode == 2) || (mode == 1)) && (ch == '-') && (cnt>0)))
              {
                if ( !( (mode == 2) && (ch == '.') && (dp) ) )
                {
                  if (!( ((mode == 1) || (mode == 2)) && (zifanz> 7) ))     // Integer und Kommazahlen auf max. 8 Ziffern beschraenken
                  {
                    *ptr= ch;
                    ptr++;
                    cnt++;
                    uart_putchar(ch);
                    if (ch== '.') { dp= 1 ; }
                    if ((ch != '.') && (ch != '-')) { zifanz++; }
                  }
                }
              }
            }
          }
        }
      }
    }
  } while ((ch != 0x0a) && (ch != 0x0d));
  *ptr= 0;

  return string;
}


/* --------------------------------------------------
                   str_cmdsplit

      zerlegt einen praeparierten Befehlsstring in
      max. 3 Bestandteile.

      String muss im Format "cmd,op1,op2" sein

      Bsp.: "wr,a,123"

      liefert: cmd = "wr"
               op1 = "a"
               op2 = "123"

      Uebergabe:
        str          : Zeiger auf Quellstring
        cmd,op1,op2 : siehe obiges Beispiel

      Rueckgabe      : Anzahl Bestandteile
   -------------------------------------------------- */
uint8_t str_cmdsplit(char *str, char *cmd, char *op1, char *op2)
{
  char *ptr;
  int  manz = 0;

  ptr = strtok(str, ",");

  while (ptr)
  {
    manz++;
    switch (manz)
    {
      case 1 : memcpy(cmd, ptr, strlen(ptr)+1); break;
      case 2 : memcpy(op1, ptr, strlen(ptr)+1); break;
      case 3 : memcpy(op2, ptr, strlen(ptr)+1); break;
      default : break;
    }
    // Test ob strtok funktioniert wie gewuenscht
//    if (ptr)  printf("%s\n\r", ptr);
    ptr = strtok(NULL, ",");
  }
  return manz;
}


uint8_t str_cmdcuts(char *src, char *cmdbuf, char *op1, char *op2)
{
  uint8_t opanz;

  // fuehrende Leerzeichen loeschen
  do
  {
    if (*src== ' ') str_delch(src,0);
  } while (*src== ' ');

  str_replacech(&src[0],',',' ');
  str_locase(&src[0]);                                         // und alles nach Kleinbuchstaben konvertieren

  // alle Leerzeichen, CR- und LF- Zeichen entfernen
  str_delallch(&src[0], 13);
  str_delallch(&src[0], 10);
  str_delallch(&src[0], ' ');
  // String liegt jetzt im Format:  "wr,a,123" vor

  op1[0]= 0; op2[0]= 0;
  opanz= str_cmdsplit(src, cmdbuf, op1, op2);                 // zerlegt Kommandozeile nach Kommando und Parameter

  return opanz;
}


/* --------------------------------------------------
                      dez2bcd

      wandelt eine dezimale Zahl in eine BCD
      Bsp: value = 45
      Rueckgabe    0x45
   -------------------------------------------------- */
uint8_t dez2bcd(uint8_t value)
{
  uint8_t hiz,loz,c;

  hiz= value / 10;
  loz= (value -(hiz*10));
  c= (hiz << 4) | loz;
  return c;
}

/* -------------------------------------------------------------
                         uc2bin

                 unsigned char to bin

     Uebergabe:
         s : Zeiger auf ein Array, das den konvertierten String
             aufnimmt. Array muss 10 Bytes gross sein (8 Zeichen
             fuer Binaerdarstellung, 1 Zeichen fuer den Nibble-
             trenner plus 1 Zeichen fuer das Stringende

     Usage:
         char pufferstr[10];
         uint8_t zahl = 0xe3;

         uc2bin(&pufferstr[0], zahl);
         printf("\n\r 0x%X = %sb \n\r", zahl, pufferstr);
   ------------------------------------------------------------- */
void uc2bin(char *s, uint8_t val)
{
  int8_t b;

  for (b= 7; b > -1; b--)
  {
    if (val & (1 << b)) { *s = '1'; }  else { *s = '0'; }
    s++;
    if (b== 4)
    {
      *s = '.';       // Punktzeichen als Trenner zwischen Hi- und Lo-Nibble
      s++;
    }
  }
  *s= 0;
}

/* --------------------------------------------------
      binout

      gibt einen Bytewert binaer auf der seriellen
      Schnittstelle aus
   -------------------------------------------------- */
void binout(uint8_t value)
{
  uint8_t b;

  for (b= 8; b != 0; b--)
  {
    if (value & (1 << (b-1))) uart_putchar('1'); else uart_putchar('0');
  }
}

/* --------------------------------------------------
                     my_readint

     liest einen Integerzahlenwert ueber die RS-232
     ein.

     Parameter:
         outtext : Zeiger auf einen String, der
                   vor einer Werteeingabe ausge-
                   geben wird.
         maxanz  : Anzahl maximal eingebarer Ziffern
   -------------------------------------------------- */
int my_readint(char *outtext, char maxanz)
{
  char txbuffer[9];

  printf("%s", outtext);
  *txbuffer= '\0';
  return my_atoi(uart_lineinput(txbuffer, 1, maxanz));	
}

/* --------------------------------------------------
     readstr

     liest einen Text ueber RS232 ein. Eine
     Korrektur mittels Backspacetaste ist moeglich

     Uebergabe:
       *buf   : Zeiger auf einen Speicherbereich, der
                den eingegebenen Text aufnimmt
       maxnz  : maximale Zeichenanzahl, die ein-
                gegeben werden kann.
   -------------------------------------------------- */
void readstr(char *buf, int maxanz)
{
  *buf = '\0';
  uart_lineinput(buf, 0, maxanz);
}

/* --------------------------------------------------------
                  Ende Helferfunktioneen
   -------------------------------------------------------- */

/* ########################################################
     Radio - Bedienerinterface (mittels UART)
   ######################################################## */

/* --------------------------------------------------
     show_tune

     zeigt die Empfangsfrequenz und die aktuell
     eingestellte Lautstaerke an
   -------------------------------------------------- */
void show_tune(void)
{
  char i;

  if (aktfreq < 1000) { my_putchar(' '); }
  printf("  %k MHz  |  Volume: ",aktfreq);

  if (aktvol)
  {
    my_putchar('0');
    for (i= 0; i< aktvol-1; i++) { my_putchar('-'); }
    my_putchar('x');
    i= aktvol;

    while (i< 15)
    {
      my_putchar('-');
      i++;
    }
  }
  else
  {
    printf("x--------------");
  }
  printf("  \r");
}

/* --------------------------------------------------
     setnewtune

     setzt eine neue Empfangsfrequenz und zeigt die
     neue Frequenz an.
   -------------------------------------------------- */
void setnewtune(uint16_t channel)
{
  aktfreq= channel;
  rda5807_setfreq(aktfreq);
  show_tune();
}


/* --------------------------------------------------
     radio_ctrl

     Bedienerinterface UKW-Radio
   -------------------------------------------------- */
void radio_ctrl(void)
{
  char ch;

  char spacestr[] = "\r                                        ";

  printfkomma= 1;                       // my_printf verwendet mit Formatter %k eine Kommastelle

  printf("\n\n\r  ----------------------------------\n\r");
  printf(      "    UKW-Radio mit I2C-Chip RDA5807\n\r");
  printf(      "  ----------------------------------\n\n\r");
  printf(      "      (+)     Volume+\n\r");
  printf(      "      (-)     Volume-\n\n\r");
  printf(      "      (u)     Empfangsfrequenz hoch\n\r");
  printf(      "      (d)     Empfangsfrequenz runter\n\r");
  printf(      "      (f)     Senderdirekteingabe\n\n\r");
  printf(      "      (1..6)  Stationstaste\n\n\r");
  printf(      "      (e)     Radio aus (Ende)\n\n\r");

  rda5807_reset();
  rda5807_poweron();
  rda5807_setmono();
  rda5807_setfreq(aktfreq);
  rda5807_setvol(aktvol);

  show_tune();

  while(1)
  {
    ch= 0;
    ch= uart_getchar();
    switch (ch)
    {
      case '+' :                            // Volume erhoehen
        {
          if (aktvol< 15)
          {
            aktvol++;
            rda5807_setvol(aktvol);
            my_putchar('\r');
            show_tune();
          }
          break;
        }

      case '-' :
        {
          if (aktvol> 0)                      // Volume verringern
          {
            aktvol--;
            rda5807_setvol(aktvol);
            my_putchar('\r');
            show_tune();
          }
          break;
        }

      case 'd' :                           // Empfangsfrequenz nach unten
        {
          if (aktfreq > fbandmin)
          {
            aktfreq--;
            setnewtune(aktfreq);
            delay(20);
            show_tune();
          }
        break;
        }
      case 'u' :                           // Empfangsfrequenz nach oben
        {
          if (aktfreq < fbandmax)
          {
            aktfreq++;
            setnewtune(aktfreq);
            delay(20);
            show_tune();
          }
          break;
        }
      case 'f' :                           // Senderdirekteingabe
        {
          printf("%s", spacestr);
          aktfreq= my_readint("\r Eingabe Freq.: ",5);
          printf("%s\r", spacestr);
          setnewtune(aktfreq);
          delay(20);
          show_tune();
          break;
        }


      case 'e' :
        {
          rda5807_setvol(0);
          rda5807_reset();
          return ;
          break;
        }

      default  : break;
    }

    if ((ch >= '1') && (ch <= '6'))
    {
      setnewtune(festfreq[ch-'0'-1]);
      show_tune();
    }
  }

}

//   Ende Radio Bedienerinterface
// -------------------------------------------------------------------------

/* --------------------------------------------------
     eep_ctrl

     Bedienerinterface EEProm
   -------------------------------------------------- */

void eep_ctrl(void)
{
  #define   txmaxbuffer  256

  char      ch;
  char      textbuf[txmaxbuffer];
  uint16_t  eep_memsize;
  uint8_t   bussize;

  do
  {
    printf("\n\n\r  ----------------------------------\n\r");
    printf(      "    EEPROM 24LCxx\n\r");
    printf(      "  ----------------------------------\n\n\r");
    printf(      "      aktuelle Deviceadresse: 0x%x\n\r",eep_addr);
    printf(      "      aktuelle Pagesize: 0x%x\n\n\r",eep_pagesize);
    printf(      "      (i)     EEProm Info\n\r");
    printf(      "      (p)     Pagesizegroesse setzen\n\r");
    printf(      "      (d)     EEProm loeschen\n\r");
    printf(      "      (w)     Text eingeben und speichern\n\r");
    printf(      "      (r)     Text aus EEProm anzeigen\n\n\r");
    printf(      "      (e)     Ende\n\r");

    ch= uart_getchar();
    switch(ch)
    {
      case 'i':
      {
        printf("\n\r Dieser Vorgang kann gespeicherte");
        printf("\n\r Daten ueberschreiben.");
        printf("\n\r Wollen sie den Vorgang fortsetzen(j/n)...");
        ch = uart_getchar();
        if( ch == 'j' || ch == 'J')
        {
          eep_memsize= eep_getmemsize();
          if (eep_memsize == 256) bussize= 8; else bussize= 16;
          printf("\n\n\r ########################################");
          printf(  "\n\r #   EEProm Info");
          printf(  "\n\r #");
          printf(  "\n\r #     Inspected device adress: 0x%x", eep_addr);

          // nachfolgende if - else if ist dem abgespeckten printf geschuldet
          // welches nur signed 16 - Bit Werte darstellen kann
          if (eep_memsize== 0xffff)
            printf(  "\n\r #     Memory-size:             65536 byte");
          else
          if (eep_memsize== 0x8000)
            printf(  "\n\r #     Memory-size:             32768 byte");
          else
            printf(  "\n\r #     Memory-size:             %d byte", eep_memsize);
          printf(  "\n\r #     Adressbus:               %d bit", bussize);
          printf(  "\n\r #");
          if (eep_memsize== 256)                                           // Speicherkapazitaet abhaengig von
                                                                           // Anzahl 256 Byte Speicherbloecken
                                                                           // (bspw. 8 Bloecke zu 256 Byte beim 24LC16
            printf("\n\r #     EEProm probably 24LC16 or less");

          if (eep_memsize== 0x1000)                                        // 4 kByte
            printf("\n\r #     EEProm probably 24LC32");
          if (eep_memsize== 0x2000)                                        // 8 kByte
            printf("\n\r #     EEProm probably 24LC64");
          if (eep_memsize== 0x4000)                                        // 16 kByte
            printf("\n\r #     EEProm probably 24LC128");
          if (eep_memsize== 0x8000)                                        // 32 kByte
            printf("\n\r #     EEProm probably 24LC256");
          if (eep_memsize== 0xffff)                                        // 64 kByte
          {
            printf("\n\r #     EEProm probably 24LC512 or ");
            printf("\n\r #     one block from a 24LC1025");
          }
          printf(  "\n\r ########################################");
        }
        printf("\n\n\r done...\n\r");
        break;
      }
      case 'd' :
      {
        printf("\n\r EEProm loeschen... ");
        eep_erase();
        printf("\n\r done...\n\r");
        break;
      }
      case 'p' :
      {
        printf("\n\r");
        eep_pagesize= my_readint(" Groesse einer Page: ",3);
        printf("\n\r");
        break;
      }
      case 'w' :
      {
        printf("\n\rText eingeben, der im EEProm gespeichert wird.");
        printf("\n\rEin Zeilenumbruch kann mit dem '#' Zeichen");
        printf("\n\reingegeben werden. Beenden der Texteingabe");
        printf("\n\rmit Enter.\n\n\r");
        readstr(&textbuf[0], 128);
        eep_writebuf(0x10, &textbuf[0], 128);
        break;
      }
      case 'r' :
      {
        eep_readbuf(0x10, &textbuf[0], 128);
        printf("\n\n\r%s\n\r", textbuf);
        ch= uart_getchar();
        break;
      }
      case 'e' :
      {
        break;
      }
      default : break;
    }
  }while (ch != 'e');

}


void i2c_scanbus(void)
{
  uint8_t   ack;
  uint16_t  cx;

  printf("\n\n\rI2C Bus scanning\n\r--------------------------\n\n\r" \
         "Devices found at address:\n\n\r");

  for (cx= 2; cx< 255; cx+= 2)
  {
    ack= i2c_start(cx);
    delay(5);
    i2c_stop();
    if (ack)
    {
      switch (cx)
      {
        case 0xC0 : printf("Adr. %xh : TEA5767 UKW-Radio\n\r", cx); break;
        case 0x20 :
        case 0x22 : printf("Adr. %xh : RDA5807 UKW-Radio\n\r", cx); break;
        case 0x40 :
        case 0x42 :
        case 0x44 :
        case 0x46 :
        case 0x48 :
        case 0x4A :
        case 0x4C :
        case 0x4E : printf("Adr. %xh : PCF8574 I/O Expander\n\r", cx); break;
        case 0x90 :
        case 0x92 :
        case 0x94 :
        case 0x96 :
        case 0x98 :
        case 0x9A :
        case 0x9C :
        case 0x9E : printf("Adr. %xh : LM75 Temp.-Sensor\n\r", cx); break;
        case 0xA0 :
        case 0xA2 :
        case 0xA4 :
        case 0xA6 :
        case 0xA8 :
        case 0xAA :
        case 0xAC :
        case 0xAE : printf("Adr. %xh : EEProm\n\r", cx); break;
        case 0xE0 :
        case 0xE2 :
        case 0xE4 :
        case 0xE6 :
        case 0xE8 :
        case 0xEA :
        case 0xEC :
        case 0xEE : printf("Adr. %xh : HT16K33 7-Segment Treiber\n\r", cx); break;
        case 0x78 : printf("Adr. %xh : SSD13016 I2C-OLED Display\n\r", cx); break;
        case 0xD0 : printf("Adr. %xh : RTC - DS1307\n\r", cx); break;

        default   : printf("Adr. %xh : unknown\n\r",cx); break;
      }
    }
    delay(1);
  }
  printf("\n\n\rEnd of I2C-bus scanning... \n\r");
  i2c_stop();

  printf("\n\r... done ! \n\r");
}

void lm75_show(void)
{
  char ch;
  int  temp, cx;

  cx= 0;
  printfkomma= 1;                       // my_printf verwendet mit Formatter %k eine Kommastelle
  printf("\n\r");
  do
  {
    ch= 0;
    if ((uart_ischar())) ch= uart_getchar();
    delay(10);

    if (!cx)
    {
      cx= 0;
      temp= lm75_read();
      printf("\r Temp.: %k [Grad Celsius]", temp);
    }
    cx++;
    cx= cx % 50;
  }while(!(ch));
  printf("\n\r");
}

void pcf8574_toggleoutput(void)
{
  static uint8_t outbyte = 0x00;
  uint8_t b, ch;

  printf("\n\r Tasten 0..7 zum ein/ausschalten der Ausgaenge");
  printf("\n\r Taste e fuer Ende\n\r");
  pcf8574_write(outbyte);
  do
  {
    ch= uart_getchar();
    if ((ch >= '0') && (ch <= '7'))
    {
      b= ch - '0';
      outbyte= outbyte ^ (1 << b);
      pcf8574_write(outbyte);
    }
  } while (ch != 'e');
}

void pcf8574_ctrl(void)
{
  char    ch;
  uint8_t ack;
  uint8_t l, lr;
  char    strbuffer[17];

  ack= i2c_start(pcf8574_addr);
  if (!(ack))
  {
    printf("\n\r PCF8574 nicht verfuegbar oder nicht angeschlossen...\n\r");
    return;
  }
  i2c_stop();

  do
  {
    printf("\n\n\r  ----------------------------------\n\r");
    printf(      "    PCF8574 I/O - Demo\n\r");
    printf(      "  ----------------------------------\n\n\r");
    printf(      "      (l)     Lauflich (LED gegen +)\n\r");
    printf(      "      (o)     Ausgaenge schalten\n\r");
    printf(      "      (i)     Pins einlesen\n\r");
    printf(      "      (e)     Ende\n\r");

    ch= uart_getchar();
    switch(ch)
    {
      case 'l' :                            // Knightrider-Lauflich
      {
        l= 0x01; lr= 0xff;
        do
        {
          ch= 0;
          if ((uart_ischar())) ch= uart_getchar();
          pcf8574_write(~l);
          if (lr) l= l << 1; else l= l >> 1;
          if (!(l))
          {
            if (lr) l= 0x40; else l= 0x02;
            lr= ~lr;
          }
          delay(100);
        }while(!(ch));
        ch= 0;
        break;
      }
      case 'i' :
      {
        printf("\n\r Einlesen Ende mit beliebiger Taste...\n\n\r");
        do
        {
          l= pcf8574_read();
          uc2bin(&strbuffer[0], l);
          printf("\r I/O als Eingaenge: 0x%x = 0b%s   ", l, strbuffer);
          uart_putchar(' ');
          delay(150);
        }while(!(uart_ischar()));
        ch= uart_getchar();
        ch= 0;
        break;
      }
      case 'o' :
      {
        pcf8574_toggleoutput();
        break;
      }
      default : break;
    }
  }while (ch != 'e');
}

/* --------------------------------------------------
     oled_txtoutput

     Das OLED-Display als "Displayschreibmaschine"
     verwenden: Jeder Tastendruck der ueber den
     UART empfangen wird, wird auf dem Display
     ausgegeben
   -------------------------------------------------- */
void oled_txtoutput(void)
{
  char ch;
  char first = 1;

  printf("\n\r Jede Tasteneingabe die getaetigt wird, wird");
  printf("\n\r auch auf dem Display ausgegeben.\n\r");
  printf("\n\r Mit '#' ist ein Zeilenumbruch moeglich,");
  printf("\n\r ein '*' loescht den Displayinhalt,");
  printf("\n\r die Entertaste beendet die Anzeige\n\n\r");

  clrscr();
  gotoxy(0,0);

  outstream= oled_stream;
  printf("Geben sie einen\n\rText ein !!!");
  outstream= uart_stream;

  do
  {
    ch= uart_getchar();
    if (first) { clrscr(); first= 0; }
    if (ch == '#')
    {
      printf("\n\r");
      oled_putchar(0x0d); oled_putchar(0x0a);
    }
    else
    {
      if (ch== '*')
      {
        clrscr();
        gotoxy(0,0);
        printf("\n\r");
      }
      else
      {
        uart_putchar(ch);
        oled_putchar(ch);
      }
    }
  }while ((ch != 0x0a) && (ch != 0x0d));
}

/* --------------------------------------------------
     oled_txtfromeep

     Text, der im EEProm gespeichert ist, auf dem
     Display anzeigen
   -------------------------------------------------- */
void oled_txtfromeep()
{
  uint8_t ack;
  char    textbuf[129];

  ack= i2c_start(eep_addr);
  if (!(ack))
  {
    printf("\n\r EEProm nicht verfuegbar oder nicht angeschlossen...\n\r");
    return;
  }
  i2c_stop();

  eep_readbuf(0x10, &textbuf[0], 128);

  outstream= oled_stream;
  ssd1306_init();
  clrscr();

  printf("%s", textbuf);
  outstream= uart_stream;
  printf("\n\r done...\n\r");
}

/* --------------------------------------------------
     oled_txtfromeep

     Zeigt Temperatur (aus LM75), Datum und einge-
     stellter Sender (aus RDA5807) auf dem Display
     an
   -------------------------------------------------- */
void oled_rtc_temp_ukw(void)
{
  struct my_datum date;

  uint8_t  is_rda, is_lm75, is_rtc;

  uint8_t abr;
  char    ch;

  char tagnam[7][3] =
  {
    "So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"
  };


  is_rda= i2c_start(rda5807_adrr << 1);
  i2c_stop();
  is_lm75= i2c_start(lm75_addr);
  i2c_stop();
  is_rtc= i2c_start(rtc_addr);
  i2c_stop();
  outstream= oled_stream;
  clrscr();
  printf("I2C-Device check\n\r---------------\n\r");
  printf("RDA5807: ");
  if (is_rda) printf("ok"); else printf("n.a.");
  printf("\n\rLM75   : ");
  if (is_lm75) printf("ok"); else printf("n.a.");
  printf("\n\rDS1307 : ");
  if (is_rtc) printf("ok"); else printf("n.a.");
  delay(2000);

  clrscr();
  printfkomma= 1;
  doublechar= 1;
  if (is_rda)
  {
    printf("%k",aktfreq);
    doublechar= 0;
    printf(" MHz");

    rda5807_reset();
    rda5807_poweron();
    rda5807_setmono();
    rda5807_setfreq(aktfreq);
    rda5807_setvol(aktvol);
  }
  else
  {
    printf("n.a.");
    doublechar= 0;
    printf("MHz");
  }
  gotoxy(0,2); printf("----------------");
  gotoxy(0,3);
  printf("Temp.:");

  abr= 0;
  while(!(abr))
  {
    if (systick_millis > 500)
    {
      systick_millis= 0;
      gotoxy(7,3);
      if (is_lm75) printf("%k %cC", lm75_read(), 0x81); else printf("n.a.");   // 0x81 Zeichen fuer hochgestelltes o

      gotoxy(3,5);
      if (is_rtc)
      {
        date= rtc_readdate();

        printf("%x.%x:%x\n\n\r",date.std, date.min, date.sek);

        printf("%s  ", tagnam[date.dow]);
        printf("%x.%x.20%x", date.tag, date.monat, date.jahr);
      }
      else
      {
        printf("RTC n.a.");
      }
    }
    if (uart_ischar())
    {
      ch= uart_getchar();
      switch (ch)
      {
        case '+' :                            // Volume erhoehen
          {
            if (aktvol< 15)
            {
              aktvol++;
              rda5807_setvol(aktvol);
            }
            break;
          }

        case '-' :
          {
            if (aktvol> 0)                      // Volume verringern
            {
              aktvol--;
              rda5807_setvol(aktvol);
            }
            break;
          }
        default  : abr= 1; break;
      }
    }

  }
  if (is_rda)
  {
    rda5807_setvol(0);
    rda5807_reset();
  }

  clrscr();
  printf("SSD1306 - OLED");

  outstream= uart_stream;
  printf("\n\r done...");
}

/* --------------------------------------------------
     oled_ctrl

     Menu zur Steuerung der Displayfunktionen
   -------------------------------------------------- */
void oled_ctrl(void)
{
  char  ch;
  uint8_t ack;

  ack= i2c_start(ssd1306_addr);
  if (!(ack))
  {
    printf("\n\r SSD1306 nicht verfuegbar oder nicht angeschlossen...\n\r");
    return;
  }
  i2c_stop();
  ssd1306_init();
  clrscr();
  gotoxy(0,0);
  outstream= oled_stream;
  printf("SSD1306 - OLED");
  outstream= uart_stream;

  do
  {
    printf("\n\n\r  ----------------------------------\n\r");
    printf(      "    SSD1306 OLED-Demo\n\r");
    printf(      "  ----------------------------------\n\n\r");
    printf(      "      (w)     Displayschreibmaschine\n\r");
    printf(      "      (t)     Text aus EEProm anzeigen\n\r");
    printf(      "      (c)     Display loeschen\n\r");
    printf(      "      (a)     Temp., Uhr und Radiofreq. anzeigen\n\n\r");
    printf(      "      (e)     Ende\n\r");

    ch= uart_getchar();
    switch(ch)
    {
      case 'w' : oled_txtoutput(); break;
      case 't' : oled_txtfromeep(); break;
      case 'c' : clrscr(); break;
      case 'a' : oled_rtc_temp_ukw(); break;
      default : break;
    }
  }while (ch != 'e');
}

/* --------------------------------------------------
      rtc_showtime

      liest das DS1307 Uhrenmodul aus und zeigt die
      Zeit an.
   -------------------------------------------------- */
void rtc_showtime(void)
{
  struct my_datum date;

  char tagnam[7][3] =
  {
    "So", "Mo", "Di", "Mi", "Do", "Fr", "Sa"
  };

  date= rtc_readdate();

  printf("\r %s  ", tagnam[date.dow]);
  printf("%x.%x.20%x  %x.%x:%x ", date.tag, date.monat, date.jahr,       \
                                  date.std, date.min, date.sek);

}


/* --------------------------------------------------
     rtc_stellen

     die Uhr benutzerabgefragt stellen
   -------------------------------------------------- */
void rtc_stellen(void)
{
  uint8_t   b;

  struct my_datum date;

  do
  {
    printf("\n\r");
    b= my_readint("Tag: ", 2);
  } while ( !((b > -1) && (b < 32)));
  b= dez2bcd(b);
  date.tag= b;

  do
  {
    printf("\n\r");
    b= my_readint("Monat: ", 2);
  } while ( !((b > -1) && (b < 13)));
  b= dez2bcd(b);
  date.monat= b;

  do
  {
    printf("\n\r");
    b= my_readint("Jahr (0..99): ",2);
  } while ( !((b > -1) && (b < 100)));
  b= dez2bcd(b);
  date.jahr= b;

  do
  {
    printf("\n\r");
    b= my_readint("Stunde: ",2);
  } while ( !((b > -1) && (b < 24)));
  b= dez2bcd(b);
  date.std= b;

  do
  {
    printf("\n\r");
    b= my_readint("Minute: ",2);
  } while ( !((b > -1) && (b < 60)));
  b= dez2bcd(b);
  date.min= b;

  do
  {
    printf("\n\r");
    b= my_readint("Sekunde: ",2);
  } while ( !((b > -1) && (b < 60)));
  b= dez2bcd(b);
  date.sek= b;

  rtc_writedate(&date);

}

/* --------------------------------------------------
     rtc_ctrl

     Bedienerinterface RTC Uhr
   -------------------------------------------------- */
void rtc_ctrl(void)
{
  char ch;

  struct my_datum date;
  uint8_t oldsek;

  ch= 0;
  oldsek= 0;

  delay(50);

  date= rtc_readdate();
  oldsek= date.sek-1;
  while(1)
  {
    printf("\n\n\r  ----------------------------------\n\r");
    printf(      "    DS1307 RTC\n\r");
    printf(      "  ----------------------------------\n\n\r");
    printf(      "      (z)     Zeige Uhrzeit\n\r");
    printf(      "      (s)     Uhr stellen\n\r");
    printf(      "      (e)     Ende\n\n\r");

    ch= uart_getchar();
    switch (ch)
    {
      case 'e' : return; break;
      case 's' : rtc_stellen(); break;
      case 'z' :
      {
        do
        {
          ch= 0;
          if ((uart_ischar())) ch= uart_getchar();
          delay(10);

          date= rtc_readdate();
          if (oldsek != date.sek)
          {
            rtc_showtime();
            oldsek= date.sek;
          }
        }while(!(ch));
        break;
      }
      default : break;
    }
  }
}

/* --------------------------------------------------
                         gethex

     liest einen Hex-Wert von der seriellen Schnitt-
     stelle ein und gibt diesen als Funtkionswert
     zurueck.
   -------------------------------------------------- */
uint8_t gethex(void)
{
  uint8_t ch;
  uint8_t value;

  do
  {
    ch= uart_getchar();
  } while (!( ((ch>= '0') && (ch<= '9')) || ((ch>= 'a') && (ch<= 'f')) ));
  uart_putchar(ch);

  if (ch > 'F')
  {
    ch= (ch-'a')+10;
  }
  else
  {
    if (ch> '9') ch= (ch-'A')+10; else ch -= '0';
  }
  value= (ch<< 4);

  do
  {
    ch= uart_getchar();
  } while (!( ((ch>= '0') && (ch<= '9')) || ((ch>= 'a') && (ch<= 'f')) ));
  uart_putchar(ch);

  if (ch > 'F')
  {
    ch= (ch-'a')+10;
  }
  else
  {
    if (ch> '9') ch= (ch-'A')+10; else ch -= '0';
  }
  value |= ch;
  return value;
}


#define cmdanz   9
const char i2c_cmds[cmdanz][9] =
{
//     1        2       3       4       5      6       7       8       9
     "start", "stop", "wr", "wrnack", "rd", "rdack", "scan", "help", "quit"
};

/* --------------------------------------------------
                      checkcmd

     vergleicht den in < string > uebergebenen Text
     mit dem gloabalen Textarray < i2c_cmds >.

     Gibt es eine Uebereinstimmung des Strings mit
     eine Element des Arrays "cmds" ,wird die
     Elementposition als Argument zurueck gegeben.
   -------------------------------------------------- */
char checkcmd(char *string)
{
  int     i,cp;
  char    ch;
  char    match;
  char    *s;

  i= 0;
  do
  {
    s= string;
    match= 1;
    cp= 0;
    do
    {
      ch= i2c_cmds[i][cp];
      if (ch!= *s) match= 0;
//      nur fuer Fehlersuchzwecke
//      printf("%d: %.2x = %c  %.2x = %c\n\r",cp,ch,ch,*s,*s);
      cp++; s++;
    }while(ch && match);
    if (match) return i+1;
    i++;
  } while (i < cmdanz);
  return 0;
}

/* --------------------------------------------------
                     helpmsg_print

     gibt Hilfenachricht zum I2C-Explorer aus
   -------------------------------------------------- */
void helpmsg_print(void)
{
  printf("\n\n\rI2C-Explorer mit CH32V003\n\r-----------------------------\n\r");
  printf("\n\rCommands are:  start   / stop    / wr      / wrnack");
  printf("\n\r               rd      / rdack   / scan    / help    / quit\n\n\r");
}


/* --------------------------------------------------
                     i2c_explorer

     Kommandozeilen-Tool zum Evaluieren eines I2C-
     Buses.
     Ueber Kommandozeileneingaben koennen Start- und
     Stopkonditionen sowie schreiben und lesen auf
     dem I2C-bus ausgefuehrt werden.
   -------------------------------------------------- */
void i2c_explorer(void)
{
  uint8_t ack;
  char    inbuffer[40];
  char    strbuffer[17];
  char    ch;

  char    cmd[20];
  char    op1[20];
  char    op2[20];
  char    error;
  uint8_t hw;

  helpmsg_print();
  while(1)
  {

    printf("\n\rI2C:> ");
    readstr(&inbuffer[0],18);

    // Eingabekommando in 3 Bestandteile zerlegen
    str_cmdcuts(&inbuffer[0], &cmd[0], &op1[0], &op2[0]);

    ch= checkcmd(&cmd[0]);
    switch(ch)
    {
      // start
      case 1 :                               // start - condition
      {
        i2c_sendstart();
        printf("\n\r  start-condition done...\n\r");
        break;
      }

      // stop
      case 2 :                               // stop - condition
      {
        i2c_stop();
        printf("\n\r  stop-condition done...\n\r");
        break;
      }

      // write
      case 3 :                               // write Data-Byte
      {
        hw= my_ahtoi(&op1[0], &error);
        if (!error)
        {
          ack= i2c_write(hw);
          printf("\n\r  writen value: 0x%x; received ack: %d\n\r", hw, ack);
        }
        else
        {
          printf("\n\r Input-Error: %s has no hexadezimal value\n\r");
        }
        break;
      }

      // write no acknowledge read
      case 4 :                               // write Data-Byte
      {
        hw= my_ahtoi(&op1[0], &error);
        if (!error)
        {
          i2c_write_nack(hw);
          printf("\n\r  writen value: 0x%x\n\r", hw);
        }
        else
        {
          printf("\n\r Input-Error: %s has no hexadezimal value\n\r");
        }
        break;
      }

      // read, no acknowledge send
      case 5 :
      {
        printf(" read, no ack sending\n\r");
        hw= i2c_read_nack();
        uc2bin(&strbuffer[0], hw);
        printf("\n\r readed value: %sb = %d = 0x%x\n\r", strbuffer, hw, hw);
        break;
      }

      // read, acknowledge send
      case 6 :
      {
        printf(" read, ack sending\n\r");
        hw= i2c_read_ack();
        uc2bin(&strbuffer[0], hw);
        printf("\n\r readed value: %sb = %d = 0x%x\n\r", strbuffer, hw, hw);
        break;
      }

      // scan
      case 7 :
      {
        i2c_scanbus();
        break;
      }

      // help
      case 8 :
      {
        helpmsg_print();
        break;
      }

      // quit
      case 9 :
      {
        printf("\n\r");
        return;
        break;                               // hier sollte das Programm nie hinkommen!
      }
      default : break;
    }

  }
}

void ht16k33_counter(void)
{
  char     ch;
  uint16_t cx;


  cx= 0;  
  do
  {
    if ((uart_ischar())) ch= uart_getchar();
  
    ht16k_setdez(cx,0,1);
    ht16k_setdp(1,1);
    delay(100);
    cx++;
    cx= cx % 10000;      
  }while(!(ch));
}

void ht16k_showtemp(void)
{
  char ch;
  int  temp, cx;

  cx= 0;
  do
  {
    ch= 0;
    if ((uart_ischar())) ch= uart_getchar();
    delay(10);

    if (!cx)
    {
      cx= 0;
      temp= lm75_read();
      ht16k_setdez(temp,1,1);
      ht16k_setdp(2,1);
      ht16k_wr2ram(bmps7asc[96],0);
    }
    cx++;
    cx= cx % 50;
  }while(!(ch));  
}

/* --------------------------------------------------
     ht16k33_ctrl

     Bedienerinterface HT16K33
   -------------------------------------------------- */

void ht16k33_ctrl(void)
{
  char      ch;  
  
  i2c_master_init();
  
  ht16k_init();
  ht16k_setdez(0,0,0);
  
  do
  {
    printf("\n\n\r  ----------------------------------\n\r");
    printf(      "    HT16K33\n\r");
    printf(      "  ----------------------------------\n\n\r");
    printf(      "      (c)     Counter\n\r");
    printf(      "      (t)     Zeige Temperatur\n\r");
    printf(      "      (e)     Ende\n\r");

    ch= uart_getchar();
    switch(ch)
    {
      case 'c' :
      {
        ht16k33_counter();
        break;
      }
      case 't' :
      {
        ht16k_showtemp();
        break;
      }
      default : break;
    }
  }while (ch != 'e');
}


/* -------------------------------------------------------------
                              main
   ------------------------------------------------------------- */
int main(void)
{
  char  ch;

  int     baud;

  SystemInit();                 // 48 MHz intern, Clock fuer PORTA, PORTC, PORTD eingeschaltet
  systick_init();
  baud= 115200;

  uart_init(baud);

  printf("\n\r--------------------------------------");
  printf("\n\rCH32V003 running at %d MHz\n\r", (uint16_t) F_CPU/1000000);
  printf("16 kByte Flash, 2 KByte RAM\n\n\r");
  printf("Baudrate: %d\n\n\r",baud);
  printf("08.04.2025  R. Seelig\n\r");
  printf("--------------------------------------\n\n\r");

  i2c_master_init();

  i2c_scanbus();

  while(1)
  {
    printf("\n\n\r  ----------------------------------\n\r");
    printf(      "    I2C - Demo\n\r");
    printf(      "  ----------------------------------\n\n\r");
    printf(      "      (l)     List I2C - Devices (scan) bus\n\r");
    printf(      "      (u)     Uhr (DS1307)\n\r");
    printf(      "      (t)     Temperatur (LM75)\n\r");
    printf(      "      (p)     EEProm (24LCxx)\n\r");
    printf(      "      (o)     SSD1306 OLED-Display\n\r");
    printf(      "      (i)     PCF8574 I/O-Expander\n\r");
    printf(      "      (r)     UKW-Radio (RDA5807) bus\n\r");
    printf(      "      (h)     HT16K33 LED 7-Segment\n\r"); 
    printf(      "      (c)     I2C-Explorer\n\n\r");

    ch = uart_getchar();
    switch(ch)
    {
      case 'l' : i2c_scanbus(); break;
      case 'r' : radio_ctrl(); break;
      case 'u' : rtc_ctrl(); break;
      case 't' : lm75_show(); break;
      case 'o' : oled_ctrl(); break;
      case 'i' : pcf8574_ctrl(); break;
      case 'p' : eep_ctrl(); break;
      case 'h' : ht16k33_ctrl(); break;
      case 'c' : i2c_explorer(); break;
      default  : break;
    }
  }
}
