/* ------------------------------------------------------------------
                              tm16xx.h

     Header zum Ansprechen eines TM16xx, Treiberbaustein fuer
     4 und 6 stellige 7-Segmentanzeigen mit gemeinsamer Anode und
     zusaetzlichem Tasteninterface.

     Geeignet fuer: TM1650, TM1651, TM1637

     MCU   : CH32V003

     Pinbelegung

       V003      7-Segment Anzeigentreiber
     --------------------------------------
        PC2                 CLK
        PC1                 DIO

     CLK und DIO sind mit jeweils 2,2 kOhm Pop-Up Wider-
     staenden nach +5V zu versehen.

     Beschreibung siehe tm16xx.dok

     13.03.2025 R. Seelig
   ------------------------------------------------------------------ */

#ifndef in_tm16xx
  #define in_tm16xx

  #include "ch32fun.h"
  #include "ch32v003_gpio.h"
  #include "tm16xx.h"

  /* ------------------------------------------------------------------
                            Controllerauswahl

       es kann (und muss) nur ein einziger Controller ausgewaehlt sein.
     ------------------------------------------------------------------ */

  #define  tm1637_ctrl              0            // gemeinsame Anode
  #define  tm1651_ctrl              0            // gemeinsame Anode
  #define  tm1650_ctrl              1            // gemeinsame Kathode

  /* ------------------------------------------------------------------
                        Auswahl einzubindender Funktionen
     ------------------------------------------------------------------ */


  #define enable_6dezdigit          0         // Code fuer 6-stellige dezimale Anzeige einbinden
  #define enable_6dezdigit_nonull   0         // Code fuer 6-stellige dezimale Anzeige mit
                                              // Ausblendung fuehrender Nullen einbinden
  #define enable_6hexdigit          0         // Code fuer 6-stellige hexadezimale Anzeige einbinden


  /* ----------------------------------------------------------
            Anschluss CLK und DIO  an den Controller
     ---------------------------------------------------------- */
  // ----------------------------------------------------------------
  //   Anmerkung zum Setzen von 1 und 0 auf den Pins
  //
  //   Die Pins sind in der Grundkonfiguration als Eingang geschaltet.
  //   Beim Setzen einer 1 wird nur die Konfiguration des Pins als
  //   Eingang benoetigt, da dieser dann hochohmig ist und die Leitung
  //   ueber den Pull-Up Widerstand auf 1 gelegt wird.
  //   Bei der Ausgabe einer 0 wird der Pin als Ausgang konfiguriert
  //   und dieser Pin mit einer 0 beschrieben
  // ----------------------------------------------------------------

  // SDA nach PC1; CLK nach PD4

  #define scl_pin        PD4
  #define sda_pin        PC1

  #define scl_init()     funPinMode( scl_pin, GPIO_CFGLR_IN_FLOAT );
  #define bb_scl_lo()    { funPinMode( scl_pin, GPIO_CFGLR_OUT_50Mhz_PP); \
                           funDigitalWrite( scl_pin, 0 ); }
  #define bb_scl_hi()    scl_init()

  #define sda_init()     funPinMode( sda_pin, GPIO_CFGLR_IN_FLOAT );
  #define bb_sda_lo()    { funPinMode( sda_pin, GPIO_CFGLR_OUT_50Mhz_PP); \
                           funDigitalWrite( sda_pin, 0 ); }
  #define bb_sda_hi()    sda_init()
  #define bb_is_sda()    funDigitalRead(sda_pin)        // wird fuer Tasteneinlesen benoetigt


/*
  Alternative Pindeklarationen ueber ch32v003_gpio.h

  #define scl_init()     PD4_float_init()
  #define bb_scl_lo()    { PD4_output_init(); PD4_clr(); }
  #define bb_scl_hi()    scl_init()

  #define sda_init()     PC1_float_init()
  #define bb_sda_lo()    { PC1_output_init(); PC1_clr(); }
  #define bb_sda_hi()    sda_init()

  #define bb_is_sda()    is_PC1()
*/


  #define puls_us        10
  #define puls_len()     delay_us(puls_us)

  /* ----------------------------------------------------------
                       Globale Variable
     ---------------------------------------------------------- */

  extern uint8_t hellig;                  // beinhaltet Wert fuer die Helligkeit (erlaubt: 0x00 .. 0x07)

  extern uint8_t  led7sbmp[16];           // Bitmapmuster fuer Ziffern von 0 .. F

  /* ----------------------------------------------------------
                           tm16_xofs
       Position, um die eine Anzeigenausgabe verschoben werden
       kann. Wird hier bspw. 2 angegeben, so erscheint die
       Ausgabe fuer Digit 0 nicht an Digit 0, sondern an
       Digit 2, Digit 1 an Digit 3 etc. Dieses dient vor allem
       dafuer, wenn ein 4-stelliger Wert auf einem 6-stelligen
       Display rechts buendig ausgegeben werden soll.
     ---------------------------------------------------------- */
   extern int8_t tm16_xofs;

  /* ----------------------------------------------------------
                           PROTOTYPEN
     ---------------------------------------------------------- */

  void tm16_start(void);
  void tm16_stop(void);
  void tm16_write (uint8_t value);
  uint8_t tm16_read(uint8_t ack);

  void tm16_init(void);
  void tm16_clear(void);
  void tm16_selectpos(int8_t nr);
  void tm16_setbright(uint8_t value);
  void tm16_setbmp(uint8_t pos, uint8_t value);
  void tm16_setzif(uint8_t pos, uint8_t zif);
  void tm16_setseg(uint8_t pos, uint8_t seg);
  void tm16_setdez(int value, uint8_t dpanz);
  void tm16_setdez_nonull(int value, uint8_t dpanz);
  void tm16_setdez2(char pos, int8_t value);
  void tm16_sethex(uint16_t value);
  uint8_t tm16_readkey(void);

  void tm16_setdez6digit(uint32_t value, uint8_t dpanz);
  void tm16_setdez6digit_nonull(int32_t value, uint8_t dpanz);
  void tm16_sethex6digit(uint32_t value);

#endif
