/* -----------------------------------------------
                    ntc7_seg.h

     einfaches Thermometer mit Ausgabe auf
     7-Segment LED-Anzeige

     NTC-Widerstand 10k Vishay NTCLE100E3-Serie
     B-Wert= 3977

     7-Segmentanzeige an TM1650

     MCU  : CH32V003
     Takt :

     15.04.2025  R. Seelig
   ----------------------------------------------- */
/*
                  CH32V003 A4M6                                        CH32V003F4P6

                  +-----------+                                      +--------------+
        PC1 / sda |1    C   16| PC0                         PD4 / a7 | 1         20 | PD3 / a4
        PC2 / scl |2    H   15| Vdd                   PD5 / a5 / utx | 2    C    19 | PD2 / a3
              PC3 |3    3   14| gnd                   PD6 / a6 / urx | 3    H    18 | PD1 / swio
         PC4 / a2 |4    2   13| PA2 / osc0 / a0           nrst / PD7 | 4    3    17 | PC7 / miso
       PC6 / mosi |5    V   12| PA1 / osc1 / a1      PA1 / osc0 / a1 | 5    2    16 | PC6 / mosi
       PC7 / miso |6    0   11| nrst / PD7           PA2 / osc1 / a0 | 6    V    15 | PC5 / sck / scl
       PD1 / swio |7    0   10| PD6 / a6 / urx                   gnd | 7    0    14 | PC4 / a2
         PD4 / a7 |8    3    9| PD5 / a5 / utx                   PD0 | 8    0    13 | PC3
                  +-----------+                                  Vdd | 9    3    12 | PC2 / scl
                                                                 PC0 | 10        11 | PC1 / sda
                                                                     +--------------+
*/
/*

   Anschlussbelegung:
 ------------------------------------------------------

     TM16xx           CH32V003
     -------------------------
       SCL   .......... PC2
       SDA   .......... PC1

   Segmentbelegung der Anzeige:

       a
      ---
   f | g | b            Segment |  a  |  b  |  c  |  d  |  e  |  f  |  g  | Doppelpunkt (nur fuer POS1) |
      ---               ---------------------------------------------------------------------------------
   e |   | c            Bit-Nr. |  0  |  1  |  2  |  3  |  4  |  5  |  6  |              7              |
      ---
       d

*/


#include <stdio.h>
#include "ch32fun.h"
#include "ch32v003_gpio.h"

#include "adc_polling.h"
#include "tm16xx.h"

/* -------------------------------------------------
     Lookup-table fuer NTC-Widerstand
     R25-Wert: 10.00 kOhm
     Pullup-Widerstand: 10.00 kOhm
     Materialkonstante beta: 3977
     Aufloesung des ADC: 10 Bit
     Einheit eines Tabellenwertes: 0.1 Grad Celcius
     Temperaturfehler der Tabelle: 0.2 Grad Celcius
   -------------------------------------------------*/
const int ntctable[] = {
  1559, 1284, 1009, 861, 759, 681, 618, 565,
  518, 476, 437, 402, 369, 337, 307, 278,
  250, 222, 195, 168, 140, 112, 84, 54,
  23, -10, -45, -85, -130, -183, -253, -360,
  -467
};

/* -------------------------------------------------
                     ntc_gettemp

    zuordnen des Temperaturwertes aus gegebenem
    ADC-Wert.

    NTC-Widerstand 10k Vishay NTCLE100E3-Serie
    B-Wert= 3977
   ------------------------------------------------- */
int ntc_gettemp(uint16_t adc_value)
{
  int p1,p2;

  // Stuetzpunkt vor und nach dem ADC Wert ermitteln.
  p1 = ntctable[ (adc_value >> 5)    ];
  p2 = ntctable[ (adc_value >> 5) + 1];

  // zwischen beiden Punkten interpolieren.
  return p1 - ( (p1-p2) * (adc_value & 0x001f) ) / 32;
}

uint8_t gradc = 0x63;             // Bitmuster fuer hochgestelltes "o"

/* -----------------------------------------------
                        main
   ----------------------------------------------- */
int main(void)
{
  uint32_t count = 0;
  int      i;
  uint16_t tgrad;

  SystemInit();                 // 48 MHz intern, Clock fuer PORTA, PORTC, PORTD eingeschaltet

  tm16_init();
  hellig= 4;
  tm16_clear();

  // Initialisieren des ADC-Wandlers,
  // Messwert hier an PC4
  adc_init(adc_on_PC4);

  while(1)
  {
    tgrad= ntc_gettemp(adc_get());
    tm16_setdez_nonull(tgrad, 2);

    // ADC etwas "beschaeftigen", hierdurch steht dann fuer die eigentliche Messung
    // der Messwert frueher stabil an    :-)
    for (i= 0; i<300; i++)
    {
      delay(3);
      tgrad= adc_get();
    }
    count++;
  }
}
