/* --------------------------------------------------------
                       pinchange_demo.c

     Demo fuer Pin-Change und Interrupt, hier an PD4

     An einem Logikwechsel von low nach high an PD4 wird
     ein Interrupt ausgeloest, dieser schaltet im Interrupt-
     handler kurz eine LED an und wieder aus.

     Hinweis: Normalerweise soll / darf in einem Interrupt
     keine delay-Funktionen aufgerufen werden, da hierdurch
     ein gesamtes System blockiert werden kann. Hier nur
     ausnahmsweise gemacht, um den Aufruf des PinChange-
     Interrupts aufzuzeigen

     MCU : ch32v003

     05.04.2025  R. Seelig
   -------------------------------------------------------- */

/*

                  CH32V003 A4M6                                        CH32V003F4P6

                  +-----------+                                      +--------------+
        PC1 / sda |1    C   16| PC0                         PD4 / a7 | 1         20 | PD3 / a4
        PC2 / scl |2    H   15| Vdd                   PD5 / a5 / utx | 2    C    19 | PD2 / a3
              PC3 |3    3   14| gnd                   PD6 / a6 / urx | 3    H    18 | PD1 / swio
              PC4 |4    2   13| PA2 / osc0 / a0           nrst / PD7 | 4    3    17 | PC7 / miso
       PC6 / mosi |5    V   12| PA1 / osc1 / a1      PA1 / osc0 / a0 | 5    2    16 | PC6 / mosi
       PC7 / miso |6    0   11| nrst / PD7           PA2 / osc1 / a1 | 6    V    15 | PC5 / sck / scl
       PD1 / swio |7    0   10| PD6 / a6 / urx                   gnd | 7    0    14 | PC4 / a2
         PD4 / a7 |8    3    9| PD5 / a5 / utx                   PD0 | 8    0    13 | PC3
                  +-----------+                                  Vdd | 9    3    12 | PC2 / scl
                                                                 PC0 | 10        11 | PC1 / sda
                                                                     +--------------+

*/

#include "ch32fun.h"
#include "ch32v003_gpio.h"

#define led_init()           PC3_output_init()
#define led_set()            PC3_set()
#define led_clr()            PC3_clr()

#define speed    100

/* --------------------------------------------------------
                      pd4_pinchange_init

     initialisiert PD4 fuer Interruptbetrieb bei auf-
     tretendem Logikwechsel, hier von low nach high
     (rising edge = positive Flanke)
   -------------------------------------------------------- */
void pd4_pinchange_init(void)
{
  RCC->APB2PCENR |= RCC_APB2Periph_AFIO;     // Takt fuer alternate Funktion GPIO einschalten
  PD4_float_init();                          // PD4 als Input schalten

  AFIO->EXTICR = AFIO_EXTICR1_EXTI4_PD;      // _EXTI4_PD  => PD4
                                             // _EXTI3_PC  => PD3 ... etc.

  EXTI->INTENR = EXTI_INTENR_MR4;            // enable extern Interrupt Line 4
                                             // _INTENR_MR3 => Line 3, waere also PD3

//  EXTI->RTENR = EXTI_FTENR_TR4;            // Trigger abfallende Flanke auf Line 4, FTE = falling trigger edge
  EXTI->RTENR = EXTI_RTENR_TR4;              // Trigger ansteigende Flanke auf Line 4, RTE = rising trigger edge
  

  NVIC_EnableIRQ( EXTI7_0_IRQn );            // enable Interrupt
}

/* --------------------------------------------------------
                    EXTI7_0_IRQHandler

     Interrupthandler fuer externen Interrupt, hier
     PinChange auf Line 4
   -------------------------------------------------------- */
void EXTI7_0_IRQHandler( void ) __attribute__((interrupt));
void EXTI7_0_IRQHandler( void )
{

  NVIC_DisableIRQ( EXTI7_0_IRQn );           // Interrupt abschalten

  led_set();      // LED kurz An zur Einschaltkontrolle
  delay(speed);
  led_clr();
  delay(speed);

  EXTI->INTFR = EXTI_Line4;                  // Interrupt quittieren
  NVIC_EnableIRQ( EXTI7_0_IRQn );            // Interrupt wieder zulassen
}

/* --------------------------------------------------------
                             main
   -------------------------------------------------------- */
int main(void)
{
  SystemInit();

  led_init();

  led_clr();
  led_set();      // LED kurz An zur Einschaltkontrolle
  delay(speed);
  led_clr();

  pd4_pinchange_init();

  while(1)             // Endlosschleife, Blinken geschieht im Interrupt
  {
    delay(1);
  }
}
