/* -----------------------------------------------
                    adc_polling.c

     Softwaremodul zum Einlesen analoger Werte
     mittels Polling (softwaregetriggertes Starten
     des ADCs)

     MCU  : CH32V003
     Takt :

     15.04.2025  R. Seelig
   ----------------------------------------------- */
/*

                  CH32V003 A4M6                                        CH32V003F4P6

                  +-----------+                                      +--------------+
        PC1 / sda |1    C   16| PC0                         PD4 / a7 | 1         20 | PD3 / a4
        PC2 / scl |2    H   15| Vdd                   PD5 / a5 / utx | 2    C    19 | PD2 / a3
              PC3 |3    3   14| gnd                   PD6 / a6 / urx | 3    H    18 | PD1 / swio
              PC4 |4    2   13| PA2 / osc0 / a0           nrst / PD7 | 4    3    17 | PC7 / miso
       PC6 / mosi |5    V   12| PA1 / osc1 / a1      PA1 / osc0 / a1 | 5    2    16 | PC6 / mosi
       PC7 / miso |6    0   11| nrst / PD7           PA2 / osc1 / a0 | 6    V    15 | PC5 / sck / scl
       PD1 / swio |7    0   10| PD6 / a6 / urx                   gnd | 7    0    14 | PC4 / a2
         PD4 / a7 |8    3    9| PD5 / a5 / utx                   PD0 | 8    0    13 | PC3
                  +-----------+                                  Vdd | 9    3    12 | PC2 / scl
                                                                 PC0 | 10        11 | PC1 / sda
                                                                     +--------------+
*/
#include "adc_polling.h"

/* -----------------------------------------------
                    adc_init

     initialisiert ADC fuer softwaregesteuertes
     Einlesen (Polling).

     Uebergabe:
        channel: Kanal, auf dem eine analoge
                 Spannung gelesen wird.

        Channel 0    <= PA2
        Channel 1    <= PA1
        Channel 2    <= PC4
        Channel 3    <= PD2
        Channel 4    <= PD3
        Channel 5    <= PD5
        Channel 6    <= PD6
        Channel 7    <= PD4
   ----------------------------------------------- */
void adc_init(uint32_t channel)
{
  uint8_t portbit;

  portbit= 1;
  // ADCCLK = 24 MHz => RCC_ADCPRE = 0: divide by 2
  RCC->CFGR0 &= ~(0x1F<<11);

  // Enable ADC, GPIO-Pins sind mit SystemInit und dort durch funGpioInitAll bereits
  //             geenabled
  RCC->APB2PCENR |= RCC_APB2Periph_GPIOD | RCC_APB2Periph_ADC1;

  switch (channel)
  {
    // Portbit wird auf Analogmode (CNF = 00) und Inputmode (MODE = 00)
    // umgeschaltet. Variable Portbit wird in Abhaengigkeit des Channels fuer
    // folgende Initialisierungen festgelegt

    case 0 :             // PA2 <= Analog in
    {
      portbit= 2;
      GPIOA->CFGLR &= ~(0xf<<(4*portbit));
      break;
    }
    case 1 :             // PA1 <= Analog in
    {
      portbit= 1;
      GPIOA->CFGLR &= ~(0xf<<(4*portbit));
      break;
    }
    case 2 :             // PC4 <= Analog in
    {
      portbit= 4;
      GPIOC->CFGLR &= ~(0xf<<(4*portbit));
      break;
    }
    case 3 :             // PD2 <= Analog in
    {
      portbit= 2;
      GPIOD->CFGLR &= ~(0xf<<(4*portbit));
      break;
    }
    case 4 :             // PD3 <= Analog in
    {
      portbit= 3;
      GPIOD->CFGLR &= ~(0xf<<(4*portbit));
      break;
    }
    case 5 :             // PD5 <= Analog in
    {
      portbit= 5;
      GPIOD->CFGLR &= ~(0xf<<(4*portbit));
      break;
    }
    case 6 :             // PD6 <= Analog in
    {
      portbit= 6;
      GPIOD->CFGLR &= ~(0xf<<(4*portbit));
      break;
    }
    case 7 :             // PD4 <= Analog-In
    {
      portbit= 4;
      GPIOD->CFGLR &= ~(0xf<<(4*portbit));
      break;
    }
    default :
    {
      portbit= channel;
      break;
    }
  }

  // ADC1-Flag an-ausschalten bewirkt Reset und initialisiert Register
  RCC->APB2PRSTR |= RCC_APB2Periph_ADC1;
  RCC->APB2PRSTR &= ~RCC_APB2Periph_ADC1;

  // Setup single conversion
  ADC1->RSQR1 = 0;
  ADC1->RSQR2 = 0;
  ADC1->RSQR3 = channel;                           // Kanaele 0..7 extern, Kanaele 8,9 intern

  // Setup sampling time
   ADC1->SAMPTR2 &= ~(ADC_SMP0<<(3 * portbit));
   ADC1->SAMPTR2 |= 7 << (3 * portbit);            // 0:7 => 3/9/15/30/43/57/73/241 cycles

  // ADC einschalten und Trigger= Softwaretrigger
  ADC1->CTLR2 |= ADC_ADON | ADC_EXTSEL;

  // Reset Kalibrierung
  ADC1->CTLR2 |= ADC_RSTCAL;
  while(ADC1->CTLR2 & ADC_RSTCAL);

  // Kalibrieren
  ADC1->CTLR2 |= ADC_CAL;
  while(ADC1->CTLR2 & ADC_CAL);
}

/* -----------------------------------------------
                    adc_get

     startet AD-Wandlung, wartet bis abgeschlossen
     und gibt den eingelesenen Wert (10-Bit)
     zurueck
   ----------------------------------------------- */
uint16_t adc_get(void)
{
  // Start AD-Wandlung
  ADC1->CTLR2 |= ADC_SWSTART;

  // warten bis fertig.
  while(!(ADC1->STATR & ADC_EOC));

  return ADC1->RDATAR;
}

/* -----------------------------------------------
                      mwert2spg

     rechnet einen vom ADC gelieferten 10-Bit
     Wert in eine Spannung um.

     Uebergabe:
       mwert : 10-Bit ADC Wert
       uref  : Spannungsreferenz in mV (die
               Spannung, die dem Maximalwert
               des ADC entspricht.

     Rueckgabe:
       Spannung in mV * 10

     Beispiel:
       mwert= 512; uref= 5000
       => 250, dieser Wert / 100 entspr.
          der Spannung (2.5V)
   ----------------------------------------------- */
uint32_t mwert2spg(uint16_t mwert, uint16_t uref)
{
  return ((mwert * 100 * uref) / 1024) / 1000;
}
