/* ------------------------------------------------------------------
                              ht16k33.c

     Softwaremodul fuer LED Anzeigentreiber HT16K33, hier fuer
     max. 8 stellige 7-Segment LED-Anzeige

     benoetigt i2c.c / i2c.h

     MCU   :  CH32V003
     Takt  :

     13.05.2025  R. Seelig
   ------------------------------------------------------------------ */

/*
                  CH32V003 A4M6                                        CH32V003F4P6
                  +-----------+                                      +--------------+
        PC1 / sda |1    C   16| PC0                         PD4 / a7 | 1         20 | PD3 / a4
        PC2 / scl |2    H   15| Vdd                   PD5 / a5 / utx | 2    C    19 | PD2 / a3
              PC3 |3    3   14| gnd                   PD6 / a6 / urx | 3    H    18 | PD1 / swio
              PC4 |4    2   13| PA2 / osc0 / a0           nrst / PD7 | 4    3    17 | PC7 / miso
       PC6 / mosi |5    V   12| PA1 / osc1 / a1      PA1 / osc0 / a0 | 5    2    16 | PC6 / mosi
       PC7 / miso |6    0   11| nrst / PD7           PA2 / osc1 / a1 | 6    V    15 | PC5 / sck / scl
       PD1 / swio |7    0   10| PD6 / a6 / urx                   gnd | 7    0    14 | PC4 / a2
         PD4 / a7 |8    3    9| PD5 / a5 / utx                   PD0 | 8    0    13 | PC3
                  +-----------+                                  Vdd | 9    3    12 | PC2 / scl
                                                                 PC0 | 10        11 | PC1 / sda
                                                                     +--------------+
                       HT16K33
                  +--------------+
              GND | 1         28 | +Vdd
        COM0 / AD | 2         27 | SDA
       COM1 / KS0 | 3    H    26 | SCL
       COM2 / KS1 | 4    T    25 | ROW0 / A2
       COM3 / KS2 | 5         24 | ROW1 / A1
             COM4 | 6    1    23 | ROW2 / A0
             COM5 | 7    6    22 | ROW3 / K1
             COM6 | 8    K    21 | ROW4 / K2
             COM7 | 9    3    20 | ROW5 / K3
ROW15 / K13 / INT | 10   3    19 | ROW6 / K4
      ROW14 / K12 | 11        18 | ROW7 / K5
      ROW13 / K11 | 12        17 | ROW8 / K6
      ROW12 / K10 | 13        16 | ROW9 / K7
       ROW11 / k9 | 14        15 | ROW10 / K8
                  +--------------+
                                                                     +--------------+
*/

#include "ht16k33.h"


uint8_t ht16k_brightness      = 15;          // Helligkeit, Werte 0..15 erlaubt

uint8_t fbuffer[bufsize];


/* --------------------------------------------------------
                       ht16k_wrcmd

     schreibt einen einzelnen Wert an die Adresse des
     Displaytreibers

     Uebergabe:
       val : zu schreibender Wert

     Rueckgabe:
       Acknowledge bei erfolgreichem Schreiben
   -------------------------------------------------------- */
uint8_t ht16k_wrcmd(uint8_t val)
{
  uint8_t ack;

  i2c_start(ht16k_addr);
  ack= i2c_write(val);
  i2c_stop();
  return ack;
}


/* --------------------------------------------------------
                       ht16k_wr2ram

     schreibt ein einzelnes Byte 'val' an die Display-
     RAM-Adresse 'segaddr'

     Uebergabe:
        val     : zu schreibender Wert
        segaddr : Adresse, an der der Wert in 'val'
                  geschickt wird.

     Rueckgabe:
        Acknowledge
   -------------------------------------------------------- */
uint8_t ht16k_wr2ram(uint8_t val, uint8_t segaddr)
{
  uint8_t ack;

  i2c_start(ht16k_addr);
  i2c_write(0x00 | segaddr);

  ack= i2c_write(val);
  i2c_stop();

  return ack;
}

/* --------------------------------------------------------
                       ht16k_wr2buffer

     schreibt den Inhalt des Framebuffers in das RAM des
     Displays.

     Da eine Commonleitung des Treibers 16 ROW-Leitungen
     betreibt, besteht ein Digit einer Commonleitung aus
     16-Bits = 2 Bytes. Aus diesem Grund belegt die
     die Adresse eines Digits im Treiber immer 2 Byte,
     so dass fuer ein 7-Segment Display immer 2
     Adressen benoetigt werden, wobei das hoeherwertige
     Byte in Verbindung mit einer max. 8 stelligen
     Anzeige somit unbenutzt ist.

     ToDo: Framebuffer auf 16-Byte erhoehen und die
     Software hier so gestalten, dass beim Schreiben
     des Buffers weitere 8 Digits an die Row-Leitungen
     8..15 angeschlossen werden koennen
   -------------------------------------------------------- */
void ht16k_wrbuffer(void)
{
  uint8_t i;

  for (i= 0; i< 8; i++)
  {
    ht16k_wr2ram(fbuffer[i], i << 1);
  }
}

/* --------------------------------------------------------
                       ht16k_setbrightness

     stellt die Helligkeit der Anzeige ein

     Uebergabe:
        val : einzustellende Helligkeit, erlaubte Werte
              sind 0x00 .. 0x0f
     Rueckgabe:
        Acknowledge
   -------------------------------------------------------- */
uint8_t ht16k_setbrightness(uint8_t val)
{
  uint8_t ack;

  ht16k_brightness= val;
  i2c_start(ht16k_addr);
  i2c_write(0xe0 | ht16k_brightness);

  ack= i2c_write(val);
  i2c_stop();
  return ack;
}

/* --------------------------------------------------------
                       ht16k_init

     initialisiert den Displaytreiber
   -------------------------------------------------------- */
void ht16k_init(void)
{
  uint8_t i;

  ht16k_wrcmd(0x21);                                // Wakeup 0x20, "Normalmode" 0x01(Oszillator an)
  ht16k_wrcmd(0x81);                                // Display 0x80, "anschalten" 0x01
  ht16k_wrcmd(0xa0);                                // Row/Int-Set 0xa0, INT as row driver 0x00
  ht16k_wrcmd(0xe0 | ht16k_brightness);

  // alle LED's loeschen
  for (i= 0; i<16; i++)
  {
    ht16k_wr2ram(0, i);
  }
}

/*  ---------------------------------------------------------
                          ht16k_setdez

      zeigt einen 32-Bit Integerwert dezimal auf der
      8-stelligen Anzeige an (beschreiben des Framebuffers
      und diesen anzeigen).

      Uebergabe:
         value  : anzuzeigender Wert
         pos    : Position, ab der der Wert rechtsbuendig
                  angezeigt wird
         nozero : 1 = fuehrende Nullen werden nicht angezeigt
                  0 = werden angezeigt
    --------------------------------------------------------- */
void ht16k_setdez(int32_t value, uint8_t pos, uint8_t nozero)
{
  int32_t teiler = 10000000;
  uint8_t i, v, b;

  memset(fbuffer, 0x00, bufsize);

  for (i= 0; i< pos; i++) teiler /= 10;
  for (i= 0; i< 8-pos; i++)
  {
    v= value / teiler;
    value= value - ( v * teiler);
    teiler /= 10;
    if ((i== (8-pos-1) && nozero)) nozero= 0;
    if (v || !nozero)
    {
      b= bmps7asc[v + zifofs];            // Bitmap der Ziffer aus Asciibitmap holen
                                          // ( zifofs ist Offset zu Ascii 0, da die ersten 32
                                          // Asciizeichen nicht implementiert sind
      fbuffer[7-i]= b;

      nozero=  0;
    }
  }
  ht16k_wrbuffer();
}

/*  ---------------------------------------------------------
                        ht16k_sethex

      zeigt einen 32-Bit Integerwert hexdezimal auf der
      8-stelligen Anzeige an (beschreiben des Framebuffers
      und diesen anzeigen).

      Uebergabe:
         value  : anzuzeigender Wert
         pos    : Position, ab der der Wert rechtsbuendig
                  angezeigt wird
         nozero : 1 = fuehrende Nullen werden nicht angezeigt
                  0 = werden angezeigt
    --------------------------------------------------------- */
void ht16k_sethex(int32_t value, uint8_t pos, uint8_t nozero)
{
  int32_t teiler = 0x10000000;
  uint8_t i, v, b;

  memset(fbuffer, 0x00, bufsize);
  for (i= 0; i< pos; i++) teiler /= 0x10;
  for (i= 0; i< 8-pos; i++)
  {
    v= value / teiler;
    value= value - ( v * teiler);
    teiler /= 0x10;
    if ((i== (8-pos-1) && nozero)) nozero= 0;
    if (v || !nozero)
    {
      if (v < 10)                                     // Bitmap der Ziffer aus Asciibitmap holen
        b= bmps7asc[v + 16];
      else
        b= bmps7asc[v + 23];

      fbuffer[7-i]= b;

      nozero=  0;

    }
  }
  ht16k_wrbuffer();
}


/*  ---------------------------------------------------------
                        ht16k_setdp

      setzt oder loescht einen Dezimalpunkt

      Uebergabe:
         pos    : Position, an der ein Dezimalpunkt gesetzt
                  oder geloescht wird
         enable : 1 = dp wird gesetzt
                  0 = dp wird geloescht
    --------------------------------------------------------- */
void ht16k_setdp(uint8_t pos, uint8_t enable)
{

  if (enable)
  {
    fbuffer[pos] |= (0x80);
  }
  else
  {
    fbuffer[pos] &= ~(0x80);
  }
  ht16k_wrbuffer();
}

/*  ---------------------------------------------------------
                        ht16k_puts

      schreibt einen String (in den Moeglichkeiten einer
      7-Segmentanzeige) in die Anzeige

      Uebergabe:
        *s  : Zeiger auf zu schreibenden String
        anz : Anzahl Digits der Anzeige
    --------------------------------------------------------- */
void ht16k_puts(char *s, uint8_t anz)
{
  anz--;
  while(*s)
  {
    ht16k_wr2ram(bmps7asc[*s-ascofs], anz * 2);
    s++;
    anz--;
  }
}


/*
    Segmentbelegung der Anzeige:

        a
       ---
    f | g | b            Segment |  a  |  b  |  c  |  d  |  e  |  f  |  g  | Doppelpunkt (nur fuer POS1) |
       ---               ---------------------------------------------------------------------------------
    e |   | c            Bit-Nr. |  0  |  1  |  2  |  3  |  4  |  5  |  6  |              7              |
       ---
        d
*/

// Bitmapmuster ASCII - Code, beginnend mit 0x20
const uint8_t bmps7asc []  =
{
// space   !     "     #     $     %     &     '
   0x00, 0x86, 0x22, 0x7E, 0x6D, 0xD2, 0x46, 0x20,     //  0 ..  7
//   (     )     *     +     ,     -     .      /
   0x29, 0x0B, 0x21, 0x70, 0x10, 0x40, 0x80, 0x52,     //  8 .. 15
//   0     1     2     3     4     5     6     7
   0x3F, 0x06, 0x5B, 0x4F, 0x66, 0x6D, 0x7D, 0x07,     // 16 .. 23
//   8     9     :     ;     <     =     >     ?
   0x7F, 0x6F, 0x09, 0x0D, 0x61, 0x48, 0x43, 0xD3,     // 24 .. 31
 //  @     A     B     C     D     E     F     G
   0x5F, 0x77, 0x7C, 0x39, 0x5E, 0x79, 0x71, 0x3D,     // 32 .. 39
//   H     I     J     K     L     M     N     O
   0x76, 0x30, 0x1E, 0x75, 0x38, 0x15, 0x37, 0x3F,     // 40 .. 47
//   P     Q     R     S     T     U     V     W
   0x73, 0x6B, 0x33, 0x6D, 0x78, 0x3E, 0x3E, 0x2A,     // 48 .. 55
//   X     Y     Z     [     \     ]     ^     _
   0x76, 0x6E, 0x5B, 0x39, 0x64, 0x0F, 0x23, 0x08,     // 56 .. 63
//    `    a     b     c     d     e     f     g
   0x02, 0x5F, 0x7C, 0x58, 0x5E, 0x7B, 0x71, 0x6F,     // 64 .. 71
//   h     i     j     k     l     m     n     o
   0x74, 0x10, 0x0C, 0x75, 0x30, 0x14, 0x54, 0x5C,     // 72 .. 79
//  p     q     r     s     t     u     v     w
   0x73, 0x67, 0x50, 0x6D, 0x78, 0x1C, 0x1C, 0x14,     // 80 .. 87
//  x     y     z     {     |     }     ~    blk
   0x76, 0x6E, 0x5B, 0x46, 0x30, 0x70, 0x01, 0x00,     // 88 .. 95
//  o
   0x63                                                // 96 ..
};

