/* ----------------------------------------------------------
                          shiftreg2.c

     Softwaremodul zur Ansteuerung der seriellen Schieber-
     egister HEF4094 oder SN74HC595 (beide Schiebe-
     register funktionieren aehnlich), haben jedoch unter-
     schiedliche Pinbelegungen)

     Primaere Funktion der Schieberegister in Verbindung mit
     Mikrocontroller ist die Erweiterung von Ausgangspins
     (GPIO - Porterweiterung)

     Dieses Softwaremodul beinhaltet u.a. Bitmuster zur
     Ansteuerung von 7-Segmentanzeigen (gemeinsame Anode
     UND gemeinsame Kathode).

     Einstellungen der Pinbelegung in shiftreg2.h

     MCU       :  CH32V003

     25.03.2025  R. Seelig
   ---------------------------------------------------------- */

/*

              HEF4094                             74HC595
          Anschlusspins IC                    Anschlusspins IC

           +------------+                      +------------+
           |            |                      |            |
       str | 1   H   16 | +Vcc            out1 | 1   H   16 | +Vcc
       dat | 2   E   15 | oe              out2 | 2   E   15 | out0
       clk | 3   F   14 | out4            out3 | 3   F   14 | dat
      out0 | 4       13 | out5            out4 | 4       13 | /oe
      out1 | 5   4   12 | out6            out5 | 5   4   12 | strobe
      out2 | 6   0   11 | out7            out6 | 6   0   11 | clk
      out3 | 7   9   10 | out_seriell     out7 | 7   9   10 | /rst
       gnd | 8   4    9 | out_seriell*     gnd | 8   4    9 | out7s
           |            |                      |            |
           +------------+                      +------------+
      * ungepufferter Ausgang

                           Anschlussbelegung

   Arduino   AVR      Pin - HEF4094           Pin - SN74HC595
   ----------------------------------------------------------------
                       16 Vcc                  16 Vcc
                       15 oe = Vcc             10 (/maser reset) = Vcc
     D11     PB3 ...    2 (seriell data)  ...  14 (seriell data)
     D7      PD7 ...    1 (strobe)        ...  12 (strobe)
     D13     PB5 ...    3 (clock)         ...  11 (clock)
                       10 (seriell out')        9 (seriell out; Q7S)
                        8 GND                   8 GND

*/

#include "shiftreg2.h"


// globale Puffervariable des Schieberegisters. Beinhaltet bei Verwendung der
// Bibliotheksfunktionen immer den aktuellen Ausgabezustands des Registers.
// Wichtig, um einzelne Bits ausgeben zu koennen, ohne bestehende Bits zu
// aendern
uint8_t sr_value = 0x00;

/* ----------------------------------------------------------
                             sr_setvalue

     transferiert einen einzeln 8-Bit Wert ins serielle
     Schieberegister (Verwendung fuer nicht kaskadierte
     Schieberegister)
   ---------------------------------------------------------- */
void sr_setvalue(uint8_t value)
{
  int8_t i;

  for (i= 7; i> -1; i--)
  {
    if ((value >> i) & 1) { srdat_set(); } else { srdat_clr(); }

    srclk_set();
    srclk_clr();                            // Taktimpuls erzeugen
  }

  srstrobe_set();                           // Strobeimpuls : Daten Schieberegister ins Ausgangslatch uebernehmen
  srstrobe_clr();
  sr_value= value;
}


/* ----------------------------------------------------------
                         sr_setkaskade

     Benutzung fuer hintereinander geschaltete (kaskadierte)
     Schieberegister (serieller Ausgang ist mit dem Daten-
     Input des nachfolgenden Schieberegisters verbunden).

     Alle Clock- und Strobeleitungen der Register sind
     parallel geschaltet.

     sr_setkaskade transferiert die in <anz> angegebene
     Anzahl Byes, die in einer uint16_t Variable
     (2 serielle Schieberegister) oder in einer uint32_t
     Variable (4 serielle Schieberegister) gespeichert
     sind.

     Ein Strobesignal erfolgt erst nach Ausgabe aller
     transferierten Bytes.

     Uebergabe
         anz    : Anzahl zu sendender Bytes
         *datas : Zeiger auf eine uint16_t oder uint32_t
                  Variable

     Benutzung:

         // Ausgabe aus uint16_t Variable
         // (2 kaskadierte Schieberegister)

         uint16_t val16 = 0x4321;
         // niederwertiges Byte wird zuerst gesendet (0x21)
         sr_setkaskade(2, (uint8_t*)&val16);

         // Ausgabe aus uint32_t Variable
         // (4 kaskadierte Schieberegister)

         uint16_t val32 = 0x87654321;
         // niederwertiges Byte wird zuerst gesendet (0x21)
         sr_setkaskade(4, (uint8_t*)&val32);

   ---------------------------------------------------------- */
void sr_setkaskade(uint8_t anz, uint8_t *datas)
{
  int8_t i, cx, value;

  if (!anz) return;
  if (anz< 2)
  {
    sr_setvalue(*datas);
    return;
  }

  for (cx= 0; cx< anz; cx++)
  {
    value = *datas;
    datas++;                                  // auf naechstes zu sendendes Element zeigen
    for (i= 7; i> -1; i--)
    {
      if ((value >> i) & 1) { srdat_set(); } else { srdat_clr(); }

      srclk_set();
      srclk_clr();                            // Taktimpuls erzeugen
    }
  }
  srstrobe_set();                             // Strobeimpuls : Daten Schieberegister ins Ausgangslatch uebernehmen
  srstrobe_clr();

}

/* -----------------------------------------------------------
     reversbyte

     Tauscht Bitwerte von LSB und MSB:
       B7 xchange B0
       B6 xchange B1
       B5 xchange B2
       B4 xchange B5
   ----------------------------------------------------------- */
uint8_t reversebyte (uint8_t x)
{
  x = ((x >> 1) & 0x55) | ((x << 1) & 0xaa);
  x = ((x >> 2) & 0x33) | ((x << 2) & 0xcc);
  x = ((x >> 4) & 0x0f) | ((x << 4) & 0xf0);
  return x;
}

/* ----------------------------------------------------------
                     sr_setarraykaskade

     Benutzung fuer hintereinander geschaltete (kaskadierte)
     Schieberegister (serieller Ausgang ist mit dem Daten-
     Input des nachfolgenden Schieberegisters verbunden).

     Alle Clock- und Strobeleitungen der Register sind
     parallel geschaltet.

     sr_setarraykaskade transferiert die in <anz> angegebene
     Anzahl Byes, die in einem unsigned char (uint8_t)
     Array in die Schieberegister und schiebt alle Bits des
     Arrays hiermit seriell aus.

     Mittels <dir> kann die Ausgaberichtung bestimmt werden.

     Bei Ausgabe des letzten Elements zuerst (dir == 1) wird
     das hoechstwertigste Bit eines Elements zuerst
     ausgeschoben.

     Bei Ausgabe des ersten Elements zuerst (dir == 0) wird
     das niederwertigste Bit eines Elements zuerst
     ausgeschoben.

     Ein Strobesignal erfolgt erst nach Ausgabe aller
     transferierten Bytes.

     Uebergabe
         anz    : Anzahl Elemente des uint8_t Arrays
         dir    : Ausgaberichtung
                  dir== 0 => erstes Element zuerst
                  dir== 1 => letztes Element zuerst
         *datas : Zeiger auf ein uint8_t Array

     Benutzung:

         #define sranz   3
         uint8_t sr_puffer[sranz] = { 0x02, 0x04, 0x01 };

         // letztes Element zuerst (0x01)
         sr_setarraykaskade(3, (uint8_t*)&sr_puffer[0]);

   ---------------------------------------------------------- */

void sr_setarraykaskade(uint8_t anz, uint8_t dir, uint8_t *datas)
{
  int8_t i, cx, value;

  if (!anz) return;
  if (anz< 2)
  {
    sr_setvalue(*datas);
    return;
  }

  if (dir== 0)
  {
    datas= datas+anz-1;
  }

  for (cx= 0; cx< anz; cx++)
  {
    if (dir== 0)
    {
      value = *(datas);
      datas--;
    }
    else
    {
      value = reversebyte(*datas);
      datas++;
    }

    for (i= 7; i> -1; i--)
    {
      if ((value >> i) &1) { srdat_set(); } else { srdat_clr(); }

      srclk_set();
      srclk_clr();                            // Taktimpuls erzeugen
    }
  }
  srstrobe_set();                             // Strobeimpuls : Daten Schieberegister ins Ausgangslatch uebernehmen
  srstrobe_clr();

}

/* ----------------------------------------------------------
                             sr_setsinglebit

     setzt ein einzelnes Ausgabnsbit im Schieberegister indem
     die Puffervariable "bitmanipuliert" wird und die Puffer-
     variable ausgegeben wird
   ---------------------------------------------------------- */
void sr_setsinglebit(char nr)
{
  sr_value |= (1 << nr);
  sr_setvalue(sr_value);
}

/* ----------------------------------------------------------
                             sr_clrsinglebit

     loescht ein einzelnes Ausgabnsbit im Schieberegister indem
     die Puffervariable "bitmanipuliert" wird und die Puffer-
     variable ausgegeben wird
   ---------------------------------------------------------- */
void sr_clrsinglebit(char nr)
{
  sr_value &= ~(1 << nr);
  sr_setvalue(sr_value);
}

