/* --------------------------------------------------------
                         systick.c

     Systemticker. Zaehlt im System automatisch 32-Bit
     Zeiteinheiten hoch:

         system_zsec    : Zehntelsekunden
         system_halfsec : togglet jede halbe Sekunde 1-0
         system_sec     : Sekunden

     einfaches Blinkprogramm: Das "Hallo Welt" fuer
     Mikrocontroller, Blinkintervall geschieht durch den
     Systemticker

     MCU : ch32v003

     04.04.2025  R. Seelig
   -------------------------------------------------------- */

/*
 * Example for using SysTick with IRQs
 * 03-25-2023 E. Brombaugh
 * 05-12-2023 C. Lohr (Modified to reflect updated sysclk)
 * 09-25-2024 ADBeta (Minor updates to main loop, comments and added
 *                    convenient macro function)
 * 04-03-2025 R. Seelig (deutsche Kommentare und eingefuegte system_halfsec, system_sec
 *                       system_zsec)
 */

#include "systick.h"

 volatile uint32_t systick_millis;
 volatile uint32_t system_zsec;
 volatile uint32_t system_halfsec;
 volatile uint32_t system_sec;


/* --------------------------------------------------------
                         systick_init

     Initialisiert den Systemticker um IRQ mit Auto-reload
     auszuloesen. Benutzt wird HCLK als Taktquelle
   -------------------------------------------------------- */
void systick_init(void)
{

  SysTick->CTLR = 0x0000;                         // reset Systemticker

  SysTick->CMP = SYSTICK_ONE_MILLISECOND - 1;     // Compare-Register um Interrupt jede Millisekunde
                                                  // auszuloesen

  SysTick->CNT = 0;                               // Systemzaehler = 0
  systick_millis = 0;                             // Systemmillisekunden = 0
  system_halfsec = 0;
  system_sec= 0;

  // Systemtickerkonfiguration setzen
  SysTick->CTLR |= SYSTICK_CTLR_STE   |  // Enable Counter
                   SYSTICK_CTLR_STIE  |  // Enable Interrupts
                   SYSTICK_CTLR_STCLK ;  // Set Clock Source to HCLK/1

  // Interrupt zulassen
  NVIC_EnableIRQ(SysTicK_IRQn);
}

/* --------------------------------------------------------
                         SysTick_Handler

     Systeminterrupt-Handler. Zaehlen der Millisekunden,
     Sekunden und toggeln der halben Sekunden
   -------------------------------------------------------- */
void SysTick_Handler(void) __attribute__((interrupt));

void SysTick_Handler(void)
{
  // Compare Register fuer naechste Interruptausloesung setzen
  SysTick->CMP += SYSTICK_ONE_MILLISECOND;

  // Triggerstatus fuer naechsten Interrupt
  SysTick->SR = 0;

  // die Zaehler
  systick_millis++;

  if ((systick_millis % 500)== 0)
  {
    system_halfsec=  system_halfsec ^ 1;
  }

  if ((systick_millis % 100)== 0)
  {
    system_zsec++;
    if ((system_zsec % 10)== 0)
    {
      system_sec++;
      system_zsec= 0;
    }
  }
}
