/* -------------------------------------------------------
                         toene.c

     erzeugt Toene der C-Dur Tonleiter mittels Timer1-
     Overflow Interrupt, Lautsprecher an PD4 (ueber
     PNP-Transistor)

     MCU   :  ch32v003

     06.05.2025  R. Seelig
   ------------------------------------------------------ */

#include "toene.h"

// Frequenztabelle der Noten, beginnend mit Frequenz fuer c''  (12 Noten per Oktave)
// 2 Oktaven gespeichert = 24 Werte

// Note (Oktave 0)            c     cis     d     dis     e      f     fis     g     gis     a     ais     h
// Frequenz [Hz]             261    277    293    311    329    349    370    392    415    440    466    494
uint16_t freqreload [24] = { 1915,  1805,  1706,  1607,  1519,  1432,  1351,  1275,  1204,  1136,  1072,  1012,

// Note (Oktave 1)            c     cis     d     dis     e      f     fis     g     gis     a     ais     h
// Frequenz [Hz]             523    554    587    622    659    698    740    784    830    880    932    988
                             958,   905,   853,   806,   762,   718,   679,   639,   606,   570,   539,   509 };


// Grundspielgeschwindigkeit
#define                 playtempo 5       // Spieltempo, in dem die Noten abgespielt werden

char    sound;

/* --------------------------------------------------------
                         notenplay_init

     initialisiert Timer1 fuer einen Compare-Match-
     Betrieb im Interruptmodus.

     Lokale Variable:

     prescale  : Taktteiler des Corecodes zum Zaehlers.
     cmp_match : Compare-Match (hier Autoreloadwert) bei
                 dem der Zaehler ueberlaeuft und einen
                 Interrupt ausloest

                 Bsp.: 48 MHz Coretakt, Prescale 48
                 bedeutet 1 MHz Timer Takt = 1 us
                 cmp_match 999 bedeutet 0..999 = 1000 Takte
                 bis zum Interruptintervall
   -------------------------------------------------------- */
void notenplay_init(void)
{
  uint16_t prescale = 47;
  uint16_t cmp_match =  freqreload[9];
  spk_init();

  RCC->APB1PCENR |= (RCC_APB1Periph_TIM2);
  TIM2->CTLR1 |= TIM_CKD_DIV1;

  TIM2->PSC= prescale;
  TIM2->ATRLR= cmp_match;

  TIM2->SWEVGR |= TIM_UG;
  TIM2->DMAINTENR |=(TIM_IT_Update);
  TIM2->CTLR1 |= TIM_CEN;

  NVIC_SetPriority(TIM2_IRQn,5);
  NVIC_EnableIRQ(TIM2_IRQn);

  sound= 0;
}


/* --------------------------------------------------------
                    TIM2_IRQHandler

      Code nach TIM_IT_UPDATE wird hier mit der Anzahl
      Mikrosekunden abgearbeitet, die in cmp_match des
      timer2_init angegeben ist
   -------------------------------------------------------- */
void TIM2_IRQHandler() __attribute__( ( interrupt() ) );

void TIM2_IRQHandler()
{
  static int toggle = 0;

  if (sound)
  {
    toggle = toggle ^ 1;
    if (toggle) { spk_set(); } else { spk_clr(); }
  }

  TIM2->INTFR &=~(uint16_t)TIM_IT_Update;
}


/* --------------------------------------------------
                       sound_off
     schaltet Ton aus
   -------------------------------------------------- */
void sound_off(void)
{
  spk_clr();
  sound= 0;
}

/* --------------------------------------------------
                       sound_on
     schaltet Ton an
   -------------------------------------------------- */
void sound_on(void)
{
  spk_clr();
  sound= 1;
  TIM2->CNT = 0;
}

/* --------------------------------------------------
                       playnote
     erzeugt eine Frequenz entsprechend der Noten-
     tabelle. Frequenz wird solange erzeugt, wie die
     globale Variable - sound - groesser Null ist.
   -------------------------------------------------- */
void playnote(uint8_t note)
{
  TIM2->ATRLR= freqreload[note];
  sound_on();
}


/* --------------------------------------------------
                          tonlen
      schaltet nach der Zeit - w - die Frequenz aus
      (und bestimmt somit, wie lange eine Note ge-
      spielt wird). Anschliessend erfolgt eine Warte-
      zeit.
   -------------------------------------------------- */
void tonlen(int w)
{
  int cx;

  w = w*100;
  for (cx= 0; cx< w; cx++) { delay_us(100); }
  sound= 0;
  delay(30);
}

/* --------------------------------------------------
                       playstring
     Interpretiert einen String als "Notenstring".
     Gueltig abspielbare Noten sind:

     c-d-e-f-g-a-h-c
         und
     C-D-F-G-A  (fuer Cis, Dis, Fis, Gis, Ais)

     Gueltige Werte fuer Notenlaengen sind:

     1-2-3-4-5-8

     Bspw. wird mit "d4" eine Viertelnote D gespielt

     + schaltet in die hoehere Oktave
     - schaltet in die untere Oktave
   -------------------------------------------------- */
void playstring(const char* s)
{
  char ch;
  char aokt;
  int dind;

  aokt= 0; dind= 0;
  ch= s[dind];
  while (ch)
  {
    ch= s[dind];
    switch(ch)
    {
      case '-': { aokt= aokt-12; break; }
      case '+': { aokt= aokt+12; break; }
      case 'c': { playnote(aokt); break; }
      case 'C': { playnote(aokt+1); break; }
      case 'd': { playnote(aokt+2); break; }
      case 'D': { playnote(aokt+3); break; }
      case 'e': { playnote(aokt+4); break; }
      case 'f': { playnote(aokt+5); break; }
      case 'F': { playnote(aokt+6); break; }
      case 'g': { playnote(aokt+7); break; }
      case 'G': { playnote(aokt+8); break; }
      case 'a': { playnote(aokt+9); break; }
      case 'A': { playnote(aokt+10); break; }
      case 'h': { playnote(aokt+11); break; }
      case '1': { tonlen(16*playtempo); break; }
      case '2': { tonlen(8*playtempo); break; }
      case '3': { tonlen(6*playtempo); break; }
      case '4': { tonlen(4*playtempo); break; }
      case '5': { tonlen(3*playtempo); break; }
      case '8': { tonlen(2*playtempo); break; }
    }
    dind++;
  }
}
