/* --------------------------------------------------------
                        uart_read.c

     Stellt Eingabefunktionen ueber die serielle Schnitt-
     stelle zur Verfuegung. Eine gemeinsame Eingabefunktion
     stellt unterschiedliche Eingaben fuer

        - Texte
        - Integerzahlen
        - Kommazahlen
        - Hexadezimale Zahlen

     bereit.
     Diese Eingabe wird in einen String eingelesen und kann
     danach in die entsprechenden Variablentypen konvertiert
     werden.

     Fuer Kommazahlen wird eine Konvertierung in die eigene
     printf-Konvention mit Ersetzungszeichen %k oder in
     ein float bereitgestellt.

     Diese Software funktioniert aufgrund des 32-Bit Integer-
     wertes nur auf 32-Bit Mikrocontrollern

     MCU :

     08.04.2025  R. Seelig
   -------------------------------------------------------- */

#include "uart_read.h"

// enum { stringinput, intinput, kommainput, hexinput };

/* --------------------------------------------------
                       strdelch

                 "String delete Char"
    loescht ein Zeichen aus einer Zeichenkette

    Uebergabe:   ptrstr  => Ursprungstring
                 pos     => Loeschposition
                            ( erstes Zeichen entspr.
                              Position 0 )
   -------------------------------------------------- */
void strdelch(char *ptrstr, int pos)
{
  int l,i;
  char *hptr;

  l= strlen(ptrstr);
  hptr= ptrstr+pos;
  for (i= pos+1; i!= l+1; i++)
  {
    *hptr= *(hptr+1);
    hptr++;
  }
}

/* --------------------------------------------------
                       upperch

                 "upper char"
    konvertiert Kleinbuchstaben a..z zu Grossbuch-
    staben A..Z

    Entspricht der Funktion "toupper" aus <string.h>
    die mit riscv-Compiler in diesem Framework
    nicht funktioniert

    Uebergabe:
      ch: zu konvertierendes Zeichen
   -------------------------------------------------- */
char upperch(char ch)
{
  if ((ch> 'a'-1) && (ch< 'z'+1))
  {
    return (ch-32);
  }
  else
  {
    return ch;
  }
}

/* --------------------------------------------------
                       strhicase

                 "String high case"
    konvertiert alle Kleinbuchstaben eines Strings
    zu Grossbuchstaben

    Uebergabe:   ptrstr  => Ursprungstring
   -------------------------------------------------- */
void strhicase(char *ptrstr)
{
  uint8_t i;

  for(i = 0; ptrstr[i]; i++)
  {
    ptrstr[i] = upperch(ptrstr[i]);
  }
}

/* ----------------------------------------------------------
                         strfindch

     sucht in einem String ein Zeichen.

     Liefert einen Zeiger auf das erste Vorkommen von
     ch in str bzw. den Wert NULL, wenn ch in str nicht
     enthalten ist.

     ==> Ersatz fuer strfindch aus <string.h> welches mit dem
     CH32FUN Framework augenscheinlich nicht richtig
     funktioniert (liefert keinen NULL-Zeiger)
   ---------------------------------------------------------- */
char *strfindch(const char *str, char ch)
{
  while(*str)
  {
    if (*str == ch) { return (char *)str; }
    str++;
  }
  if ((*str == 0) && (ch== '\0')) return (char *)str;
  return NULL;
}

#if (( enable_int_convert == 1) || (enable_komma_convert == 1) || (enable_float_convert == 1))

  /* ----------------------------------------------------------
                            my_pow10

       einfache Funktion 10 hoch x
     ---------------------------------------------------------- */
  int my_pow10(unsigned int x)
  {
    #define my_powbase    10

    int i;
    int result = my_powbase;

    if (x== 0)
    {
      return 1;
    }
    else
    {
      for (i = 1; i < x; i++)
      {
        result *= my_powbase;
      }
      return result;
    }
  }

  /* ----------------------------------------------------------
                            my_atoi

                       my ascii to integer

       einfache Konvertierung eines Strings zu einem Integer-
       wert in Ermangelung eines atoi

       Uebergabe:
          *str: Zeiger, der einen im String enthaltenen
                Integerzahlenwert repraesentiert

       Rueckgabe:
          konvertierter Integerwert
     ---------------------------------------------------------- */
  int my_atoi(char *str)
  {
    int i, j;
    int res;
    int sign;
    int len;

    i = 0;
    res = 0;
    sign = 1;                        // Vorzeichenflag, 1 = positiv, -1 = negativ

    len= strlen(str);

    while(str[i] == ' ' || (str[i] >= 9 && str[i] <= 13))
    {
      i++;
    }

    if(str[i] == '-')
    {
      sign = -1;
      i++;
    }
    else if(str[i] == '+')
    {
      sign = 1;
      i++;
    }

    j = i-1;                                // Index auf Stringanfang

    i = len;
    while (str[i] < '0' || str[i] > '9')    // Suche nach der letzten gueltigen Ziffer
    {
      i--;
    }
    int num_end = i;

    // Konvertierung mit der niederwertigsten Stelle beginnen
    while(i > j)
    {
      if (str[i] >= '0' && str[i] <= '9')
      {
        res += my_pow10(num_end-i) * (str[i] - '0');
      } else
      {
        num_end--;                          // ueberspringen einer ungueltigen Ziffer
      }
      i--;
    }
    return(res * sign);
  }

  #if ((enable_komma_convert == 1) || (enable_float_convert == 1))

    /* ----------------------------------------------------------
                              my_aftok

                        my ascii float to komma

         Konvertiert eine Kommazahl (keine wissenschaftliche
         Darstellung) in eine Kommazahl, indem sie das Komma
         aus dem String entfernt und in Kommapos die Position
         des Kommapunktes zurueck gibt.

         Uebergabe:
            *str: Zeiger, der einen im String enthaltenen
                  Hexadezimalwert repraesentiert

            *kommapos: an der im konvertieren Wert das Festkomma
                       einzufuegen ist

         Rueckgabe:
            konvertierter Integerwert
       ---------------------------------------------------------- */
    int my_aftok(char *str, char *kommapos)
    {
      char *kptr;
      char temptxt[15];
      char kpos, kp;
      int  val;

      strcpy(temptxt,str);
      kptr= strfindch(str, '.');
      kpos= kptr-str;
      if (kptr != NULL)
      {
        strdelch(temptxt, kpos);
        kp= strlen(temptxt) - kpos;
      }
      else
      {
        kp= 0;
      }
      val= my_atoi(temptxt);

      *kommapos= kp;
      return val;
    }
  #endif

  #if (enable_float_convert == 1)
    /* ----------------------------------------------------------
                              my_aftof

                        my ascii float to float

         Konvertiert eine Kommazahl (keine wissenschaftliche
         Darstellung) in einen float.

         Uebergabe:
            *str: Zeiger, der einen im String enthaltenen
                  Hexadezimalwert repraesentiert

            *kommapos: an der im konvertieren Wert das Festkomma
                       einzufuegen ist

         Rueckgabe:
            konvertierter Integerwert
       ---------------------------------------------------------- */
    float my_aftof(char *str)
    {
      char  kp;
      int   val;
      float f;

      val= my_aftok(str, &kp);

      f= val;
      if (kp)
      {
        do
        {
          f /= 10.0;
          kp--;
        } while(kp);
      }
      return f;
    }
  #endif

#endif

/* ----------------------------------------------------------
                          my_ahtoi

                    my ascii hex to integer

     einfache Konvertierung eines Hex-Strings zu einem
     Integerwert

     Uebergabe:
        *str: Zeiger, der einen im String enthaltenen
              Hexadezimalwert repraesentiert

     Rueckgabe:
        konvertierter Integerwert
   ---------------------------------------------------------- */
int my_ahtoi(char *str)
{
  char     ch;
  uint32_t result;

  strhicase(str);

  result= 0;
  do
  {
    ch= *str;
    if ((ch < '0') || ((ch > '9') && (ch < 'A')) || (ch > 'F') )
    {
      return 0xffffffff;
    }
    if (ch< 'A') { ch -= '0'; } else { ch -= 55; }
    result = (result << 4) + ch;

    str++;
  }  while (*str);

  return result;
}

/* ------------------------------------------------------------
                            uart_puts

     gibt einen AsciiZ String auf dem UART aus
  -------------------------------------------------------------  */
void uart_puts(char *p)
{
  while(*p)
  {
    uart_putchar(*p);
    p++;
  }
}


/* --------------------------------------------------
                      uart_lineinput

     Eingabe eines Strings in den Buffer *string

     Beim Start der Funktion wird der in *string ab-
     gelegte Text angezeigt und der Cursor an das
     Ende dieses Textes gesetzt.

     Einschraenkungen:

     Im Integer- und Kommamodus ist die Eingabe von
     Ziffern auf max. 8 begrenzt (da es sonst zu
     einem Werteueberlauf kommen kann).


     Uebergabe:

            *string   : Zeiger auf String, der zum
                        einen angezeigt wird, zum
                        anderen wird in diesen Speicher-
                        bereich auf den der Zeiger
                        zeigt der String abgelegt

            mode      : Eingabemodus
                           0 => Stringeingabe
                           1 => Integereingabe, nur
                                Ziffern von 0..9,'-'
                           2 => Kommazahleingabe
                                 wie Mode 1 mit
                                zusaetzlich erlaubtem
                                '.' Zeichen
                           3 => Hexadezimaleingabe,
                                Eingabe von 0..9, A..F

            anz       : max. Anzahl Zeichen, die
                        eingegeben werden koennen

     Rueckgabe:
       Zeiger auf den verarbeiteten String
   -------------------------------------------------- */
char *uart_lineinput(char *string, uint8_t mode, char anz)
{

  uint8_t      cnt;
  char         ch;
  char         *ptr;
  char         *kptr;
  char         dp, tc;
  char         zifanz;

  dp= 0;
  if (strlen(string)> 0)
  {
    kptr = strfindch(string, '.');
    if (kptr) dp= 1;
  }

  if (mode== 3) { strhicase(string); }

  // Ziffernanzahl fuer Integer und Komma auf 8 begrenzen, daher feststellen
  // wieviele Ziffern im Vorgabestring enthalten sind
  if ((mode== 1) || (mode== 2))
  {
    zifanz= strlen(string);
    if (zifanz > 0)
    {
      if (dp) { zifanz--; }
      if (*string == '-') { zifanz--; }
    }
//    printf("\n\rZifanz: %d\n\r",zifanz);
  }

  cnt= strlen(string);
  uart_puts(&string[0]);
  ptr= string+cnt;

  do
  {
    ch= uart_getchar();

    if ((ch!= 0x0a) && (ch != 0x0d))                // Enter Taste ?
    {

      if ((ch== 8) || (ch== 127))                   // Backspace
      {
        if (cnt> 0)                                 // ist ueberhaupt ein Zeichen vorhanden
        {
          tc= *(ptr-1);
          if (tc == '.') { dp= 0; }
          if ( !((tc == '.') || (tc == '-'))) { zifanz--; }
          ptr--;
          cnt--;
          *ptr= 0;                                  // neues Stringende

          // letzes Zeichen im Terminal loeschen
          uart_putchar(0x08);
          uart_putchar(' ');
          uart_putchar(0x08);
        }
      }
      else
      {                                                                         // Zeicheneingabe, erlaubte Zeichen in den entsprechenden
                                                                                // Modi pruefen
        if (mode==3)
        {
          if ((ch>= 'a') && (ch<= 'f')) { ch -= 32; }                           // bei hexadezimaler Eingabe Grossbuchstaben verwenden
        }


        if (  ((mode== 0) && (ch >= 32) && (ch <= 126))                         // anhaengen Ascii-Zeichen (Stringmode)

           || (( (mode== 1) && (ch >= '0') && (ch <= '9') )  || (ch== '-'))     // anhaengen Dezimalziffer und '-' Zeichen (Integermode)

           || ((mode== 2) && (((ch >= '0') && (ch <= '9')) ||                   // anhaengen Kommazahl (Floatmode)
                               (ch=='.') || (ch=='-')))

           || ((mode== 3) && ( ((ch >= '0') && (ch <= '9')) ||                  // anhaengen Hexadezimalzifer (Hexmode)
                               ((ch >= 'A') && (ch <= 'F'))) ) )

        {
          if ((cnt< anz) && (!((mode==3) && (ch== '-'))) )
          {
            // '-' Zeichen nur erlauben im Integer- und Floatmode und auch nur, wenn noch kein
            // Minuszeichen angegeben wurde und nur, wenn der Cursor ganz am Anfang steht
            if (!( ((mode == 2) || (mode == 1)) && (ch == '-') && (cnt>0)))
            {
              if ( !( (mode == 2) && (ch == '.') && (dp) ) )
              {
                if (!( ((mode == 1) || (mode == 2)) && (zifanz> 7) ))     // Integer und Kommazahlen auf max. 8 Ziffern beschraenken
                {
                  *ptr= ch;
                  ptr++;
                  cnt++;
                  uart_putchar(ch);
                  if (ch== '.') { dp= 1 ; }
                  if ((ch != '.') && (ch != '-')) { zifanz++; }
                }
              }
            }
          }
        }
      }
    }
  } while ((ch != 0x0a) && (ch != 0x0d));
  *ptr= 0;

  return string;
}
