/* --------------------------------------------------------
                         uart_demo.c

     Demoprogramm fuer die serielle Schnittstelle

     TxD an PD5, RxD an PD6

     MCU : ch32v003

     25.03.2025  R. Seelig
   -------------------------------------------------------- */

/*
                  CH32V003 A4M6                                        CH32V003F4P6

                  +-----------+                                      +--------------+
        PC1 / sda |1    C   16| PC0                         PD4 / a7 | 1         20 | PD3 / a4
        PC2 / scl |2    H   15| Vdd                   PD5 / a5 / utx | 2    C    19 | PD2 / a3
              PC3 |3    3   14| gnd                   PD6 / a6 / urx | 3    H    18 | PD1 / swio
              PC4 |4    2   13| PA2 / osc0 / a0           nrst / PD7 | 4    3    17 | PC7 / miso
       PC6 / mosi |5    V   12| PA1 / osc1 / a1      PA1 / osc0 / a0 | 5    2    16 | PC6 / mosi
       PC7 / miso |6    0   11| nrst / PD7           PA2 / osc1 / a1 | 6    V    15 | PC5 / sck / scl
       PD1 / swio |7    0   10| PD6 / a6 / urx                   gnd | 7    0    14 | PC4 / a2
         PD4 / a7 |8    3    9| PD5 / a5 / utx                   PD0 | 8    0    13 | PC3
                  +-----------+                                  Vdd | 9    3    12 | PC2 / scl
                                                                 PC0 | 10        11 | PC1 / sda
                                                                     +--------------+
*/

#include <stdio.h>
#include <stdlib.h>

#include "ch32fun.h"
#include "ch32v003_gpio.h"

#include "uart.h"
#include "my_printf.h"
#include "term_colpos.h"

#define printf       my_printf


#define led_init()               PC3_output_init()
#define led_set()                PC3_set()
#define led_clr()                PC3_clr()

#define inp_init()               PD4_input_init()
#define is_inp()                 is_PD4()


#define cnt_speed    500

#define M_PI         3.14159265359f

#define baud         115200


/* ----------------------------------------------------------
                        my_putchar

     wird von my_printf aufgerufen. Hierauf gibt my_printf
     den Datenstream aus
   ---------------------------------------------------------- */
void my_putchar(char ch)
{
  uart_putchar(ch);
}


// Lookup-Tabelle fuer Sinuswinkel (DEG-Mode) von 0..90 Grad multipliziert
// mit 1000 => Integerwerte mit 3 Nachkommastellen.
// Schrittweite Arrayelemente 1 Grad

uint16_t sindeg_lookup[91] =
{
     0,  17,  34,  52,  69,  87, 104, 121,
   139, 156, 173, 190, 207, 224, 241, 258,
   275, 292, 309, 325, 342, 358, 374, 390,
   406, 422, 438, 453, 469, 484, 500, 515,
   529, 544, 559, 573, 587, 601, 615, 629,
   642, 656, 669, 681, 694, 707, 719, 731,
   743, 754, 766, 777, 788, 798, 809, 819,
   829, 838, 848, 857, 866, 874, 882, 891,
   898, 906, 913, 920, 927, 933, 939, 945,
   951, 956, 961, 965, 970, 974, 978, 981,
   984, 987, 990, 992, 994, 996, 997, 998,
   999, 999, 1000
};


/* ----------------------------------------------------------
                          sindeg

     liefert fuer einen Sinuswinkel (DEG-Mode) den Sinuswert
     multipliert mit 1000 => Integerwert mit 3 Nachkommastellen

     Uebergabe:

         w : Winkel

     Rueckgabe:
         Sinuswert * 1000

     Bsp.:  x= sindeg(45);

            x== 707 entspricht 0.707
   ---------------------------------------------------------- */
int sindeg(int w)
{
  int sign= 1;

  w= w % 360;

  if (w> 180)
  {
    sign= -1;
    w -= 180;
  }
  if (w> 90) { w= 180-w; }
  return (sindeg_lookup[w] * sign);
}


/* ----------------------------------------------------------
                        draw_sincurve

     zeichnet eine Ascii-Sinuskurve
   ---------------------------------------------------------- */
void draw_sincurve(void)
{
  int      i, x, x2;
  uint8_t  col;

  col= 0;
  for (i= 0; i< 361; i += 10)
  {
    textcolor((col+9));
    col++;
    col= col % 7;

    x2 = ((sindeg(i) / 42) + 25);
    for (x= 0; x < x2; x++) uart_putchar(' ');
    printf("o\n\r");
    delay(40);
  }
  printf("\n\r");
}


/* ----------------------------------------------------------
                              main
   ---------------------------------------------------------- */
int main(void)
{
  int         li;
  uint16_t    counter = 0;
  uint16_t    ch;
  int         wuwerti;

  SystemInit();                 // 48 MHz intern, Clock fuer PORTA, PORTC, PORTD eingeschaltet

  li= F_CPU / 1000000;

  led_init();

  uart_init(baud);

  settextattr(0x07);
  clrscr();
  settextattr(0x1e);
  printf("\n\r --------------------------------------\n\r");
  printf(" CH32V003 running at %d MHz            \n\r", (uint16_t)li);
  printf(" 16 kByte Flash, 2 KByte RAM           \n\r");
  printf("                                       \n\r");
  printf(" Baudrate: %d                      \n\r",baud);
  printf("                                       \n\r");
  printf(" 25.03.2025  R. Seelig                 \n\r");
  printf(" --------------------------------------\n\n\r");
  gotoxy(15,12);

  textcolor(black);
  bkcolor(grey);
  printf(" Demo fuer readint ");
  textcolor(grey);
  bkcolor(black);
  printf("\n\n\r Counter preset eingeben: ");
  counter= readint();
  printf("\n\n\r Counter startet bei: %d\n\n\r", counter);

  printf("Taste um Counter anzuhalten bzw. neu zu starten\n\n\r");

  while(uart_ischar()) {ch= uart_getchar(); my_putchar(ch); }  // eventuell eingegangene Zeichen alle loeschen

  printfkomma= 3;

  while (1)
  {
    led_set();
    delay(cnt_speed);
    led_clr();

    wuwerti= sindeg(counter);

    printf("  Counter: %xh = %dd | sin(%d)= %k | PD4= ", counter, counter, counter % 360, wuwerti);

    if (is_inp()) printf("1       \r");
             else printf("0       \r");

    delay(cnt_speed);
    counter++;
    counter = counter % 3600;

    if (uart_ischar())
    {
      ch= uart_getchar();
      printf("\n\n\rGedrueckte Taste war: %c\n\r", ch );
      printf("Beliebige Taste fuer Counterstart...\n\n\r");
      ch= uart_getchar();
      draw_sincurve();
      textcolor(grey);
      bkcolor(black);
    }
  }
}
