/* --------------------------------------------------------
                         uart_receive_irq.c

     Demoprogramm fuer die serielle Schnittstelle

     TxD an PD5, RxD an PD6

     MCU : ch32v003

     10.04.2025  R. Seelig
   -------------------------------------------------------- */

/*
                  CH32V003 A4M6                                        CH32V003F4P6

                  +-----------+                                      +--------------+
        PC1 / sda |1    C   16| PC0                         PD4 / a7 | 1         20 | PD3 / a4
        PC2 / scl |2    H   15| Vdd                   PD5 / a5 / utx | 2    C    19 | PD2 / a3
              PC3 |3    3   14| gnd                   PD6 / a6 / urx | 3    H    18 | PD1 / swio
              PC4 |4    2   13| PA2 / osc0 / a0           nrst / PD7 | 4    3    17 | PC7 / miso
       PC6 / mosi |5    V   12| PA1 / osc1 / a1      PA1 / osc0 / a0 | 5    2    16 | PC6 / mosi
       PC7 / miso |6    0   11| nrst / PD7           PA2 / osc1 / a1 | 6    V    15 | PC5 / sck / scl
       PD1 / swio |7    0   10| PD6 / a6 / urx                   gnd | 7    0    14 | PC4 / a2
         PD4 / a7 |8    3    9| PD5 / a5 / utx                   PD0 | 8    0    13 | PC3
                  +-----------+                                  Vdd | 9    3    12 | PC2 / scl
                                                                 PC0 | 10        11 | PC1 / sda
                                                                     +--------------+
*/

#include <stdio.h>
#include <stdlib.h>

#include "ch32fun.h"
#include "ch32v003_gpio.h"
#include "uart.h"

#include "my_printf.h"

#define baud           4800

#define printf         my_printf

#define led_init()     PC3_output_init()
#define led_set()      PC3_set()
#define led_clr()      PC3_clr()

#define led_speed    500

// -----------------------------------------------------------
//   Ringpuffer fuer serielle Schnittstelle USART
// -----------------------------------------------------------
#define RECV_BUF_SIZE          256    // Puffergroesse Ringspeicher
uint8_t recv_buf[RECV_BUF_SIZE];

volatile int recv_buf_next;           // naechste Adresse an der eingegangenes Zeichen
                                      // abgelegt wird
volatile int recv_buf_current;	      // Adresse, an der das naechste Zeichen zu lesen ist

#define recv_buf_not_empty()          (recv_buf_current != recv_buf_next)



/* ----------------------------------------------------------
                        my_putchar

     wird von my_printf aufgerufen. Hierauf gibt my_printf
     den Datenstream aus
   ---------------------------------------------------------- */
void my_putchar(char ch)
{
  uart_putchar(ch);
}

void inp_n_delay(uint16_t cx)
{
  uint8_t ch;

  while(cx)
  {
    if (uart_ischar())
    {
      ch= uart_getchar();
      uart_putchar(ch);
    }
    cx--;
    delay(1);
  }
}

/* ----------------------------------------------------------
                        USART1_IRQHandler

     Interruptvektor, wird aufgerufen, wenn auf der UART
     ein Zeichen eingegangen ist !
   ---------------------------------------------------------- */
void USART1_IRQHandler(void) __attribute__((interrupt));
void USART1_IRQHandler(void)
{
  int cx;             // Timeoutzaehler
  int i;

  {
    cx= 0;
    do
    {
      if (USART1->STATR & USART_STATR_RXNE)           // wenn ein eingegangenes Zeichen den Interrupt
                                                      // ausgeloest hat
      {
        // wenn ein Zeichen empfangen wurde, dieses im Ringspeicher ablegen
        recv_buf[recv_buf_next] = USART1->DATAR;

        // Test auf Ueberlauf
        i = (recv_buf_next + 1) % RECV_BUF_SIZE;
        if (i != recv_buf_current)
        {
          recv_buf_next = i;
        }
      }
      cx++;
    } while (((USART1->STATR & USART_STATR_RXNE) != 0) && (cx< 10000));
  }
}

/* --------------------------------------------------------
                      buffer_getchar

     liest ein Zeichen aus dem Ringspeicher. Ist keines
     eingetroffen, wird auf den Eingang eines Zeichens
     auf der USART gewartet.
   -------------------------------------------------------- */
uint8_t buffer_getchar(void)
{
  uint8_t   ch = 0;

  while (recv_buf_current == recv_buf_next);

  if (recv_buf_current != recv_buf_next)
  {
    ch = recv_buf[recv_buf_current];
    recv_buf_current= (recv_buf_current + 1) % RECV_BUF_SIZE;
  }
  return ch;
}

/* ----------------------------------------------------------
                              main
   ---------------------------------------------------------- */
int main(void)
{
  uint8_t ch;

  SystemInit();                 // 48 MHz intern, Clock fuer PORTA, PORTC, PORTD eingeschaltet

  uart_init(baud);

  NVIC_EnableIRQ(USART1_IRQn);            // Interrupt USART1 zulassen
  USART1->CTLR1 |= USART_CTLR1_RXNEIE;    // Flag Receive-Interrupt zulassen

  led_init();

  printf("\n\r Ringbufferdemo mit Interrupt fuer");
  printf("\n\r eingegangene Zeichen \n\n\r");

  while(1)
  {
    // solange Zeichen im Ringbuffer vorhanden sind, werden diese gelesen
    // und zurueck gesendet. Nach dem Senden erfolgt absichtlich eine
    // "eigentliche" Blockade des Programmflusses mittels led_set und delay
    // um aufzuzeigen, dass waehrend dieser Zeit der Interrupt die Zeichen
    // auf dem UART entgegen nimmt
    while (recv_buf_not_empty())          // solange Zeichen im Ringbuffer sind
    {                                     // diese zuruecksenden
      ch= buffer_getchar();
      uart_putchar(ch);
    }

    led_set();
    delay(led_speed);
    led_clr();
    delay(led_speed);
  }
}
